package org.lightj.concurrent;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.FutureTask;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * thread pool executor that takes {@link IQueueTask}
 * 
 * @author biyu
 *
 */
public abstract class AbstractThreadPoolExecutor extends ThreadPoolExecutor {
	
	private static ThreadLocal<Long> executionTime = new ThreadLocal<Long>() {
		protected synchronized Long initialValue() {
			return Long.valueOf(System.nanoTime());
		}
	};

	/** executor is paused */
	protected boolean isPaused;
	
	/** pause lock */
	protected ReentrantLock pauseLock = new ReentrantLock();
	
	/** pause condition */
	protected Condition unpaused = pauseLock.newCondition();

	/** internal task queue */
	protected CompositeBlockingQueue<Runnable> queue;
	
	/** stats */
	protected ThreadPoolStats threadPoolStats;

	/**
	 * constructor
	 *
	 */
	AbstractThreadPoolExecutor(int corePoolSize,
			int maximumPoolSize,
            long keepAliveTime,
            TimeUnit unit,
            CompositeBlockingQueue<Runnable> workQueue,
            ThreadFactory threadFactory) 

    {
		super(corePoolSize,	maximumPoolSize, keepAliveTime, unit, workQueue, threadFactory);
		this.queue = workQueue;

	}

	@Override
    public Future<?> submit(Runnable task) {
        if (task == null) throw new NullPointerException();
        IQueueTask qTask = (task instanceof IQueueTask) ? (IQueueTask) task : new RunnableQueueTask(task);
        FutureTask<Object> ftask = new FutureTask<Object>((Runnable) qTask, null);
        execute(new FutureQueueTask<Object>(qTask, ftask));
        return ftask;
    }

	@Override
    public <T> Future<T> submit(Runnable task, T result) {
        if (task == null) throw new NullPointerException();
        IQueueTask qTask = (task instanceof IQueueTask) ? (IQueueTask) task : new RunnableQueueTask(task);
        FutureTask<T> ftask = new FutureTask<T>((Runnable) qTask, result);
        execute(new FutureQueueTask<T>(qTask, ftask));
        return ftask;
    }

	@SuppressWarnings("unchecked")
	@Override
    public <T> Future<T> submit(Callable<T> task) {
        if (task == null) throw new NullPointerException();
        IQueueTask qTask = (task instanceof IQueueTask) ? (IQueueTask) task : new CallableQueueTask<T>(task);
        FutureTask<T> ftask = new FutureTask<T>((Callable<T>) qTask);
        execute(new FutureQueueTask<T>(qTask, ftask));
        return ftask;
    }

	
	/**
	 * increment activity counter before execution
	 */
	protected void beforeExecute(Thread t, Runnable r) {
		getQueueStats().adjustWaitTime(((IQueueTask)r).getStats().waitEnd());
		pauseLock.lock();
		try {
			while (isPaused) unpaused.await();
		} catch (InterruptedException ie) {
			t.interrupt();
		} finally {
			pauseLock.unlock();
		}
		IQueueTask qtask = (IQueueTask) r;
		qtask.getGroup().getStats().incExecutingCounter();
		executionTime.get();
		super.beforeExecute(t, r);
	}
	
	/**
	 * decrement activity counter after execution
	 */
	protected void afterExecute(Runnable r, Throwable t) {
		try {
			long executionTimeNano = System.nanoTime() - executionTime.get(); 
			threadPoolStats.incTotalCount();
			queue.getStats().incTotalCount();
			IQueueTask qtask = (IQueueTask) r;
			qtask.getGroup().getStats().decExecutingCounter();
			qtask.getGroup().getStats().adjustExecutionDuration(executionTimeNano);
			qtask.getGroup().getStats().incTotalCount();
			super.afterExecute(r, t);
		} finally {
			executionTime.remove();
		}
	}
	
	/**
	 * Executes the given task sometime in the future.
	 */
	public void execute(Runnable command) {
		((IQueueTask) command).getStats().waitStart();
		super.execute(command);
	}

	/**
	 * pause the executor
	 *
	 */
	public final void pause() {
		pauseLock.lock();
		try {
			isPaused = true;
		} finally {
			pauseLock.unlock();
		}
	}

	/**
	 * resume the executor
	 *
	 */
	public final void resume() {
		pauseLock.lock();
		try {
			isPaused = false;
			unpaused.signalAll();
		} finally {
			pauseLock.unlock();
		}
	}
	
	/** if the threadpool is paused */
	public final boolean isPaused() {
		return isPaused;
	}







	/** pool stat */
	final ThreadPoolStats getStats() {
		return threadPoolStats;
	}
	
	/** set pool stat */
	final void setStats(ThreadPoolStats stats) {
		this.threadPoolStats = stats;
	}
	
	/** get Queue stats */
	final QueueStats getQueueStats() {
		return queue.getStats();
	}

	/**
	 * wrapper class for underneath queue
	 * 
	 * @author biyu
	 *
	 * @param <V>
	 */
	public static class FutureQueueTask<V> implements Future<V>, IQueueTask, Runnable {
		private FutureTask<V> futureTask;
		private IQueueTask queueTask;
		FutureQueueTask(IQueueTask queueTask, FutureTask<V> futureTask) {
			this.futureTask = futureTask;
			this.queueTask = queueTask;
		}
		public boolean cancel(boolean mayInterruptIfRunning) {
			return futureTask.cancel(mayInterruptIfRunning);
		}
		public V get() throws InterruptedException, ExecutionException {
			return futureTask.get();
		}
		public V get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
			return futureTask.get(timeout, unit);
		}
		public boolean isCancelled() {
			return futureTask.isCancelled();
		}
		public boolean isDone() {
			return futureTask.isDone();
		}
		public void run() {
			futureTask.run();
		}
		public QueueTaskGroup getGroup() {
			return queueTask.getGroup();
		}
		public String getName() {
			return queueTask.getName();
		}
		public QueueTaskPriority getPriority() {
			return queueTask.getPriority();
		}
		public QueueTaskStatus getStatus() {
			return queueTask.getStatus();
		}
		public QueueTaskStats getStats() {
			return queueTask.getStats();
		}
		public int compareTo(IQueueTask o) {
			return queueTask.compareTo(o);
		}
		public String toString() {
			return queueTask.toString();
		}
		public IQueueTask getQueueTask() {
			return queueTask;
		}
	}

}
