package org.lightj.concurrent;

import java.util.Collection;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * a fifo queue that support blocking and dump based on size and time limitations
 * can be used in queuing work for batch processing
 * 
 * @author biyu
 *
 * @param <E>
 */
public class BatchedFIFOBlockingQueue<E> {

    /** Lock held by take, poll, etc */
    private final ReentrantLock batchLock = new ReentrantLock();

    /** Wait queue for waiting takes */
    private final Condition full = batchLock.newCondition();
    
    /** queue not empty */
    private final Condition notEmpty = batchLock.newCondition();
    
    /** internal fifo queue */
    private final LinkedBlockingQueue<E> queue;

    /** how long before the queue gets dumped */
	private final long maxBatchTime;
	
	/** how long before the queue gets dumped time unit */
	private final TimeUnit maxBatchTimeUnit;
	
	/** how many elements before the queue gets dumped */
	private final int maxBatchSize;
	
	/** nano time when the first element gets inserted */
	private volatile AtomicLong firstInsertTimeNano = new AtomicLong(0);
	
	/** constructor */
	public BatchedFIFOBlockingQueue(long maxBatchTime, TimeUnit maxTimeUnit, int maxBatchSize) {
		this.maxBatchTime = maxBatchTime;
		this.maxBatchTimeUnit = maxTimeUnit;
		this.maxBatchSize = maxBatchSize;
		this.queue = new LinkedBlockingQueue<E>();
	}

	/** drain all elements in the queue to a collection, or wait if queue is empty */
	public int drainTo(Collection<? super E> c) throws InterruptedException {
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
        	if (queue.isEmpty()) {
        		notEmpty.await();
        	}
        	long delta = checkTimer();
        	if (delta > 0) {
           		try {
					full.await(delta, maxBatchTimeUnit);
				} catch (InterruptedException e) {
					// shouldn't have happend, ignore but restore the interrupt state so it can be propagated up
					Thread.currentThread().interrupt();
				}
        	}
        	return queue.drainTo(c);
        } finally {
        	batchLock.unlock();
        }
	}

	/** add an element in the tail of the queue */
	public boolean offer(E o) {
		boolean rst;
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
        	if (queue.isEmpty()) {
        		rst = queue.offer(o);
        		if (rst) {
        			signalNotEmpty();
        		}
        	}
        	else {
        		rst = queue.offer(o);
        		if (rst && queue.size() >= maxBatchSize) {
        			signalFull();
        		}
        	}
        } finally {
        	batchLock.unlock();
        }
		return rst;
	}
	
	public int getMaxBatchSize() {
		return maxBatchSize;
	}

	public long getMaxBatchTime() {
		return maxBatchTime;
	}

	public TimeUnit getMaxBatchTimeUnit() {
		return maxBatchTimeUnit;
	}

	/** wake up a consumer that waits on an empty queue */
	private void signalNotEmpty() {
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
       		resetTimer();
       		notEmpty.signal();
        } finally {
        	batchLock.unlock();
        }
	}

	/** wake up a consumer that waits for batch condition reached */ 
	private void signalFull() {
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
        	full.signal();
        } finally {
        	batchLock.unlock();
        }
	}
	
	/** reset time when first element enqueued */
	private void resetTimer() {
		firstInsertTimeNano.set(System.nanoTime());
	}
	
	/** check timer and return delta to wait */
	private long checkTimer() {
       	long delta = firstInsertTimeNano.getAndSet(0);
       	return delta > 0 ? TimeUnit.NANOSECONDS.convert((System.nanoTime() - maxBatchTime), maxBatchTimeUnit) : 0L; 
	}

}
