package org.lightj.concurrent;

import java.util.Collection;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

public class BatchedLockableBlockingQueue<E> {
	
    /** Lock held by take, poll, etc */
    private final ReentrantLock batchLock = new ReentrantLock();

    /** queue not empty */
    private final Condition notEmpty = batchLock.newCondition();
    
    /** put all cosumers in waiting */
    private volatile boolean consumerLocked;
    private final Condition consumerLock = batchLock.newCondition();
    
    /** internal queue */ 
	private final BlockingQueue<E> queue;
	
	/** constructor */
	public BatchedLockableBlockingQueue(BlockingQueue<E> queue) {
		this.queue = queue;
	}

	/** drain all elements in the queue to a collection, or wait if queue is empty */
	public int drainTo(Collection<? super E> c) throws InterruptedException {
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
        	if (queue.isEmpty()) {
        		notEmpty.await();
        	}
        	if (consumerLocked) {
        		consumerLock.await();
        	}
        	return queue.drainTo(c);
        } finally {
        	batchLock.unlock();
        }
	}

	/** add an element in the tail of the queue */
	public boolean offer(E o) {
		boolean rst;
        final ReentrantLock batchLock = this.batchLock;
        int c = -1;
        batchLock.lock();
        try {
        	c = queue.size();
       		rst = queue.offer(o);
        } finally {
        	batchLock.unlock();
        }
        if (c == 0) signalNotEmpty();
		return rst;
	}
	
	/** release the lock on consumer */
	public void release() {
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
        	consumerLocked = false;
        	consumerLock.signal();
        } finally {
        	batchLock.unlock();
        }
	}
	
	/** lock the consumer */
	public void lock() {
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
        	consumerLocked = true;
        } finally {
        	batchLock.unlock();
        }
	}

	/** wake up a consumer that waits on an empty queue */
	private void signalNotEmpty() {
        final ReentrantLock batchLock = this.batchLock;
        batchLock.lock();
        try {
       		notEmpty.signal();
        } finally {
        	batchLock.unlock();
        }
	}
	
	/** size of the queue */
	public int size() {
		return queue.size();
	}
}
