package org.lightj.concurrent;

import java.util.AbstractQueue;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A blocking queue consists of three parts:
 * Preemptive part
 * Round Robin part
 * Low priority (Can Wait) part
 * 
 * Elements are inserted into different part based on their {@link IQueueTask#getType()}
 * Elements are polled out in the order of preemptive, round-robin, low priority
 *  
 * @author biyu
 *
 */
public class PriorityRoundRobinBlockingQueue extends AbstractQueue<Runnable> 
	implements BlockingQueue<Runnable> 
{
	
	/** size of the queue */
    private transient int size = 0;
	
    /** Main lock guarding all access */
    private final ReentrantLock lock;

    /** Condition for waiting takes */
    private final Condition notEmpty;

    /** preemptive part of the queue */
	private LinkedList<Runnable> preemptiveQueue;
    
	/** round robin part of the queue */
	private RoundRobinPriorityQueue roundRobinQueue;
    
	/** can wait part of the queue */
	private LinkedList<Runnable> canwaitQueue;
	
	/**
	 * constructor
	 * 
	 * @param qmonitor
	 */
	public PriorityRoundRobinBlockingQueue() {
        lock = new ReentrantLock();
        notEmpty = lock.newCondition();
        preemptiveQueue = new LinkedList<Runnable>();
        roundRobinQueue = new RoundRobinPriorityQueue();
        canwaitQueue = new LinkedList<Runnable>();
	}

	@Override
	/**
	 * Use it with care, it takes a complete new snapshot of the existing data
	 * and create a new collection, this can be expensive if the queue is large
	 * also the iterator doesn't support remove from iterator 
	 */
	public Iterator<Runnable> iterator() {
		// taking a snapshot of the underneath queue and creating a new data structure
		// in order to prevent CurrentModificationException
		List<Runnable> snapshot = new ArrayList<Runnable>();
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
    		snapshot.addAll(preemptiveQueue);
    		snapshot.addAll(roundRobinQueue);
    		snapshot.addAll(canwaitQueue);
        } finally {
            lock.unlock();
        }
        Iterator<Runnable> iter = snapshot.iterator();
        return new SnapshotIterator<Runnable>(iter);
	}

	@Override
	/**
	 * size of the queue, 
	 */
	public int size() {
		return size;
	}

	/**
	 * calculate size from datas structures underneath
	 * @return
	 */
	private int calcSize() {
		int size = 0;
  		size += preemptiveQueue.size();
  		size += roundRobinQueue.size();
  		size += canwaitQueue.size();
  		return size;
	}

	/**
	 * drain all elements in this queue to a collection
	 */
	public int drainTo(Collection<? super Runnable> c) {
        if (c == null)
            throw new NullPointerException();
        if (c == this)
            throw new IllegalArgumentException();
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
    		c.addAll(preemptiveQueue);
    		c.addAll(roundRobinQueue);
    		c.addAll(canwaitQueue);
    		preemptiveQueue.clear();
    		roundRobinQueue.clear();
    		canwaitQueue.clear();
    		return c.size();
        } finally {
        	this.size = calcSize();
            lock.unlock();
        }
	}

	/**
	 * drain up to maxElements number of elements in this queue to a collection.
	 * Note the order from round robin queue is not deterministic among groups
	 */
	public int drainTo(Collection<? super Runnable> c, int maxElements) {
        if (c == null)
            throw new NullPointerException();
        if (c == this)
            throw new IllegalArgumentException();
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
        	int count = 0;
        	for (Iterator<Runnable> iter = preemptiveQueue.iterator(); iter.hasNext();) {
        		if (count++ <= maxElements) {
            		c.add(iter.next());
            		iter.remove();
        		}
        	}
        	for (Iterator<Runnable> iter = roundRobinQueue.iterator(); iter.hasNext();) {
           		if (count++ <= maxElements) {
               		c.add(iter.next());
               		iter.remove();
           		}
        	}
        	for (Iterator<Runnable> iter = canwaitQueue.iterator(); iter.hasNext();) {
        		if (count++ <= maxElements) {
            		c.add(iter.next());
            		iter.remove();
        		}
        	}
        	return count;
        } finally {
        	this.size = calcSize();
            lock.unlock();
        }
	}
	
	/**
	 * Inserts the specified element at the tail of one of the internal queues based on its {@link QueueTaskType}
	 * Note: not thread safe, has to be guarded externally
	 * @param o
	 * @return
	 */
	private boolean insert(Runnable o) {
		QueueTaskType type = ((IQueueTask) o).getGroup().getType();
		boolean rst = false;
		switch (type.getType()) {
		case 1:
			rst = preemptiveQueue.offer(o);
			break;
		case 2:
			rst = roundRobinQueue.offer(o);
			break;
		case 3:
			rst = canwaitQueue.offer(o);
			break;
		default:
			// task type not supported, don't insert
			rst = false;
		}
		if (rst) {
			this.size++;
			notEmpty.signal();
		}
		return rst;
	}

	/**
	 * unbounded
	 */
	public int remainingCapacity() {
		return Integer.MAX_VALUE;
	}

	/**
	 * Retrieves and removes the head of this queue, waiting if necessary up to the specified wait time if no elements are present on this queue.
	 */
	public Runnable poll(long timeout, TimeUnit unit)
			throws InterruptedException {
        final ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            long nanos = unit.toNanos(timeout);
            for (;;) {
                if (size() != 0) {
                    Runnable x = extract();
                    return x;
                }
                if (nanos <= 0)
                    return null;
                try {
                    nanos = notEmpty.awaitNanos(nanos);
                } catch (InterruptedException ie) {
                    notEmpty.signal(); // propagate to non-interrupted thread
                    throw ie;
                }

            }
        } finally {
            lock.unlock();
        }
	}

	/**
	 * Retrieves and removes the head of this queue, waiting if no elements are present on this queue.
	 */
	public Runnable take() throws InterruptedException {
        final ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        try {
            try {
                while (size() == 0)
                    notEmpty.await();
            } catch (InterruptedException ie) {
                notEmpty.signal(); // propagate to non-interrupted thread
                throw ie;
            }
            Runnable x = extract();
            return x;
        } finally {
            lock.unlock();
        }
	}

	/**
	 * Retrieves and removes the head of this queue, or null  if this queue is empty.
	 */
	public Runnable poll() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
        	return extract();
        } finally {
            lock.unlock();
        }
	}
	
	/**
	 * extract an element from one of the internal queues based on their ordering
	 * Note: not thread safe, has to be guarded externally
	 * @return
	 */
	private Runnable extract() {
		Runnable task = null;
		task = preemptiveQueue.poll();
		if (task == null) {
			task = roundRobinQueue.poll();
		}
		if (task == null) {
			task = canwaitQueue.poll();
		}
		if (task != null) {
			this.size--;
		}
		return task;
	}

	/**
	 * Retrieves, but does not remove, the head of this queue, returning null if this queue is empty.
	 */
	public Runnable peek() {
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
    		Runnable task = null;
    		task = preemptiveQueue.peek();
    		if (task == null) {
    			task = roundRobinQueue.peek();
    		}
    		if (task == null) {
    			task = canwaitQueue.peek();
    		}
    		return task;
        } finally {
            lock.unlock();
        }
	}

	/**
	 * to be consistent with {@link BlockingQueue} interface that takes {@link Runnable}
	 * the queue internally only support {@link IQueueTask}
	 */
	public boolean offer(Runnable o) {
        if (o == null) throw new NullPointerException();
        final ReentrantLock lock = this.lock;
        lock.lock();
        try {
        	return insert(o);
        } finally {
            lock.unlock();
        }
	}

	/**
	 * to be consistent with {@link BlockingQueue} interface that takes {@link Runnable}
	 * the queue internally only support {@link IQueueTask}
	 * this queue is always boundless, it will never need the timeout
	 */
	public boolean offer(Runnable o, long timeout, TimeUnit unit) throws InterruptedException {
		return offer(o);
	}

	/**
	 * to be consistent with {@link BlockingQueue} interface that takes {@link Runnable}
	 * the queue internally only support {@link IQueueTask}
	 */
	public void put(Runnable o) throws InterruptedException {
		offer(o);
	}
	
	void setRoundRobinQueueAlgorithm(RoundRobinQueueAlgorithm<Runnable> algorithm) {
		
	}
	
	/** round robin part of the queue */
	RoundRobinPriorityQueue getRoundRobinQueue() {
		return roundRobinQueue;
	}

    
	static class SnapshotIterator<E> implements Iterator<E> {
		Iterator<E> iter;
		SnapshotIterator(Iterator<E> iter) {
			this.iter = iter;
		}

		public boolean hasNext() {
			return iter.hasNext();
		}

		public E next() {
			return iter.next();
		}

		public void remove() {
			throw new UnsupportedOperationException("Remove is not supported");
		}
		
	}
}
