package org.lightj.concurrent;

import java.util.Collection;
import java.util.Collections;
import java.util.PriorityQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.lightj.util.StringUtil;


/**
 * Immutable QueueTaskGroup
 * 
 * @author biyu
 *
 */
public final class QueueTaskGroup {
	
	/** all groups */
	private static ConcurrentMap<String, QueueTaskGroup> groups = new ConcurrentHashMap<String, QueueTaskGroup>();
	
	/** circular array for round robin access, used for out */
	static CircularArray<PriorityQueue<Runnable>> rrArrays = new CircularArray<PriorityQueue<Runnable>>();

	/** queue map for all groups, used for in */
	static ConcurrentMap<QueueTaskGroup, PriorityQueue<Runnable>> rrQueues = new ConcurrentHashMap<QueueTaskGroup, PriorityQueue<Runnable>>();

	/** create a new instance */
	static QueueTaskGroup newInstance(String groupName, QueueTaskType type) {
		QueueTaskGroup group = new QueueTaskGroup(groupName, type);
		groups.putIfAbsent(groupName, group);
		if (type == QueueTaskType.roundrobin) {
			PriorityQueue<Runnable> queue = new PriorityQueue<Runnable>(100);
			rrArrays.add(queue);
			rrQueues.putIfAbsent(group, queue);
		}
		return group;
	}

	/** all registered groups */
	public static Collection<QueueTaskGroup> getAllGroups() {
		return Collections.unmodifiableCollection(groups.values());
	}
	
	/** task groupname */
	private String groupName;
	/** task type */
	private QueueTaskType type;
	/** stats */
	private QueueTaskGroupStats stats;
	
	/** constructor */
	private QueueTaskGroup(String groupName, QueueTaskType type) {
		if (StringUtil.isNullOrEmptyAfterTrim(groupName)) {
			throw new IllegalArgumentException("groupName cannot be empty");
		}
		this.groupName = groupName;
		this.type = type;
		this.stats = new QueueTaskGroupStats(groupName);
	}

	/** get group name */
	public String getGroupName() {
		return groupName;
	}

	/** get type */
	public QueueTaskType getType() {
		return type;
	}
	
	/** hashcode */
	public int hashCode() {
		return groupName.hashCode();
	}
	
	/** equals */
	public boolean equals(Object o) {
		if (o instanceof QueueTaskGroup) {
			return this.groupName.equals(((QueueTaskGroup)o).groupName);
		}
		return super.equals(o);
	}

	/** get stats */
	public QueueTaskGroupStats getStats() {
		return stats;
	}
	
	/**
	 * find an existing group by name
	 * @param groupName
	 * @return
	 */
	public static QueueTaskGroup getGroupByName(String groupName) {
		return groups.containsKey(groupName) ? groups.get(groupName) : QueueTaskGroups.UNKNOWN;
	}
}
