package org.lightj.concurrent;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import org.lightj.util.nanoxml.XMLElement;


/**
 * collect group stats
 * 
 * @author biyu
 *
 */
public final class QueueTaskGroupStats extends CounterStats {

	/** executing count, tasks that are actively being executed */
	private final AtomicInteger executingCount = new AtomicInteger(0);
	
	/** average execution duration in ms */
	private long avgExecutionDuration = 0L;

	/** worst execution duration in ms */
	private long worstExecutionDuration = 0L;
	
	/** resetable execution count */
	private final AtomicInteger nonresetTotal = new AtomicInteger(0);
	
	/** constructor */
	public QueueTaskGroupStats(String name) {
		super(name, "QueueTaskGroupStats");
	}
	
	/** increment total execution counter */
	public final long incCount() {
		nonresetTotal.incrementAndGet();
		return super.incTotalCount();
	}
	
	/** increment executing counter by 1 */
	public final int incExecutingCounter() {
		return executingCount.incrementAndGet();
	}
	
	/** decrement executing counter by 1 */
	public final int decExecutingCounter() {
		return executingCount.decrementAndGet();
	}
	
	/** executing counter */
	public final int getExecutingCount() {
		return executingCount.get();
	}

	/**
	 * recalculate execution durations (worst and mean)
	 * @param executionDuration
	 */
	public synchronized void adjustExecutionDuration(long executionDuration) {
		avgExecutionDuration += ((executionDuration-avgExecutionDuration)/(getTotalCount()+1));
		worstExecutionDuration = Math.max(worstExecutionDuration, executionDuration);
	}
	
	/** average duraction */
	public long getAvgExecutionDuration() {
		return avgExecutionDuration;
	}

	/** worst duration */
	public long getWorstExecutionDuration() {
		return worstExecutionDuration;
	}

	/** reset all stats */
	public synchronized void reset() {
		super.reset();
		avgExecutionDuration = 0;
		worstExecutionDuration = 0;
	}
	
	@Override
	protected void asXml(XMLElement statElem) {
		super.asXml(statElem);
		statElem.setAttribute("totalCount", nonresetTotal.get());
		statElem.setAttribute("executingCount", executingCount.get());
		statElem.setAttribute("averageExecutionDuration", TimeUnit.MILLISECONDS.convert(avgExecutionDuration, TimeUnit.NANOSECONDS));
		statElem.setAttribute("worstExecutionDuration", TimeUnit.MILLISECONDS.convert(worstExecutionDuration, TimeUnit.NANOSECONDS));
	}

}
