package org.lightj.config;

import java.lang.reflect.Field;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Category;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * base configurations
 * 
 * @author biyu
 * 
 */
public abstract class BaseRuntimeConfigs implements IRuntimeConfigs {

	/** logger */
	static Category logger = Logger.getLogger(BaseRuntimeConfigs.class);

	/**
	 * listener map, config name to listener map
	 */
	private Set<ConfigEventListener> eventListeners = new HashSet<ConfigEventListener>();

	/**
	 * add a listener, when a config field (typically a static field) changes, be notified
	 * 
	 * @param fieldName
	 * @param mylistener
	 */
	public synchronized void addListener(ConfigEventListener mylistener) {
		Class<? extends ConfigEventListener> klass = mylistener.getClass();
		boolean exist = false;
		for (ConfigEventListener listener : eventListeners) {
			if (klass == listener.getClass()) {
				exist = true;
				break;
			}
		}
		if (!exist) {
			eventListeners.add(mylistener);
		}
	}

	/**
	 * remove a listener from listener map
	 * 
	 * @param fieldName
	 * @param listener
	 */
	public synchronized void removeListener(ConfigEventListener listener) {
		eventListeners.remove(listener);
	}
	
	/**
	 * compare and set a configuration value based on config field name (typically a static field)
	 * 
	 * @param fieldName
	 * @param newValue
	 * @throws ConfigUpdateException
	 */
	public void updateValue(String fieldName, Object newValue) throws ConfigUpdateException {
		try {
			this.updateValue(this.getClass().getField(fieldName), fieldName, newValue);
		} catch (SecurityException e) {
			logger.log(Level.ERROR, e);
			throw new ConfigUpdateException(e);
		} catch (NoSuchFieldException e) {
			logger.log(Level.ERROR, e);
			throw new ConfigUpdateException(e);
		}
	}
	
	/**
	 * compare and set a configuration value based on config {@link Field}
	 * 
	 * @param field
	 * @param cfgName
	 * @param newValue
	 * @throws ConfigUpdateException
	 */
	public void updateValue(Field field, String cfgName, Object newValue) throws ConfigUpdateException {
		try {

			Object oldValue = field.get(this);
			// check if both are null, then equal, otherwise not equal
			if (newValue == null && oldValue == null) {
				// both are equal so no change
			} 
			else if (newValue == null || oldValue == null) {
				// reset
				setConfigValue(field, oldValue, newValue);
			} 
			else if (!newValue.equals(oldValue)) {
				// change
				setConfigValue(field, oldValue, newValue);
			} 
			else {
				// noop
			}

		} catch (IllegalArgumentException e) {
			logger.log(Level.ERROR, e);
			throw new ConfigUpdateException(e);
		} catch (SecurityException e) {
			logger.log(Level.ERROR, e);
			throw new ConfigUpdateException(e);
		} catch (NoSuchFieldException e) {
			logger.log(Level.ERROR, e);
			throw new ConfigUpdateException(e);
		} catch (NullPointerException e) {
			logger.log(Level.ERROR, e);
			throw new ConfigUpdateException(e);
		} catch (IllegalAccessException e) {
			logger.log(Level.ERROR, e);
			throw new ConfigUpdateException(e);
		}
	}
	
	/**
	 * set value
	 * @param configName
	 * @param oldValue
	 * @param newValue
	 * @throws SecurityException
	 * @throws NoSuchFieldException
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 */
	private void setConfigValue(Field field, Object oldValue, Object newValue)
		throws SecurityException, NoSuchFieldException,	IllegalArgumentException, IllegalAccessException 
	{
		field.set(this, newValue);
		String fieldName = field.getName();
		notifyConfigEventListener(ConfigEventEnum.CHANGE, fieldName, oldValue, newValue);
	}

	/**
	 * notify listener
	 * @param event
	 * @param fieldName
	 * @param oldValue
	 * @param newValue
	 */
	protected void notifyConfigEventListener(ConfigEventEnum event, String fieldName, Object oldValue, Object newValue) 
	{
		for (ConfigEventListener listener : eventListeners) {
			if (listener.listenToEvent().contains(event) && 
					listener.listenToFields().contains(fieldName)) {
				listener.onConfigEvent(event, fieldName, oldValue, newValue);
				logger.debug("Event called for" + fieldName);
			}
		}
	}
}
