package org.lightj.config;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.InetAddress;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.lang.StringEscapeUtils;
import org.hsqldb.lib.StringInputStream;
import org.lightj.RuntimeContext;
import org.lightj.util.Log4jProxy;
import org.lightj.util.StringUtil;
import org.lightj.util.nanoxml.XMLElement;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Manage configuration from configuration xml file
 * 
 * @author biyu
 * 
 */
@SuppressWarnings({"rawtypes","unchecked"})
public class ConfigManager implements IConfigManager {

	/** logger */
	static Log4jProxy cat = Log4jProxy.getInstance(ConfigManager.class.getName());

	/**
	 * Constants
	 */
	public static final String NODE_SET = "set";
	public static final String NODE_STRING = "string";
	public static final String NODE_BOOLEAN = "boolean";
	public static final String NODE_INT = "int";
	public static final String NODE_LIST = "list";
	public static final String NODE_MAP = "map";
	public static final String NODE_FLOAT = "float";

	/** products */
	static ConcurrentMap<String, Wrapper> products = new ConcurrentHashMap<String, Wrapper>();

	/**
	 * internal properties cache, hierarchical
	 */
	private HashMap propsMap = new HashMap();

	/**
	 * all properties in a flat map, for displaying purposes only
	 */
	private HashMap flatMap = null;
	
	/** 
	 * path to config file 
	 */
	protected URL path;
	
	/** 
	 * product 
	 */
	protected String product;
	
	/**
	 * env
	 */
	protected EnvironmentEnum env;
	
	/**
	 * ConfigManager instance by product name
	 * @param productName
	 * @return
	 */
	public static synchronized final ConfigManager getInstanceByProduct(String productName) {
		if (products.containsKey(productName)) {
			return products.get(productName).mgr;
		}
		return null;
	}
	
	/**
	 * init using a dummy config
	 * @param configRoot
	 * @param productName
	 * @return
	 */
	public static synchronized final ConfigManager init(URL configRoot, String productName) {
		return init(configRoot, productName, new DummyRuntimeConfigs());
	}
	
	/**
	 * 
	 * @param configRoot
	 * @param productName
	 * @param configs
	 * @return
	 */
	public static synchronized final ConfigManager init(URL configRoot, String productName, IRuntimeConfigs configs) {
		ConfigManager mgr = new ConfigManager(configRoot, productName);
		mgr.load();
		configs.load(mgr);
		products.putIfAbsent(productName, new Wrapper(mgr, configs));
		return mgr;
	}
	
	/**
	 * load from xml string passed in
	 * @param xmlString
	 * @param configs
	 * @return
	 */
	public static synchronized final ConfigManager init(String xmlString, IRuntimeConfigs configs) 
	{
		ConfigManager mgr = new ConfigManager(null, null);
		mgr.load(xmlString);
		configs.load(mgr);
		return mgr;
	}

	/**
	 * refresh a product
	 * @param product
	 */
	public static synchronized final void refresh(String product) {
		if (products.containsKey(product)) {
			Wrapper w = products.get(product);
			w.mgr.load();
			w.configs.load(w.mgr);
		}
		else {
			throw new IllegalArgumentException("Unknown product " + product);
		}
	}

	/**
	 * constructor
	 */
	protected ConfigManager(URL path, String product) {
		this.path = path;
		this.product = product;
	}

	/**
	 * derive an {@link EnvironmentEnum} from hostname
	 * 
	 * @return
	 */
	public final EnvironmentEnum deriveEnvironment() {
		
		EnvironmentEnum env = EnvironmentEnum.NONE;

		// try to get it from property file
		try {
			String host = InetAddress.getLocalHost().getHostName();
			String hostPropFilename = product + "." + host.toUpperCase() + ".props.xml";
			DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
			domFactory.setNamespaceAware(true); // never forget this!
			DocumentBuilder builder = domFactory.newDocumentBuilder();
			Document doc = builder.parse(new URL(path, hostPropFilename).openStream());

			XPathFactory factory = XPathFactory.newInstance();
			XPath xpath = factory.newXPath();
			XPathExpression expr = xpath.compile("//set[@name='environment']/string[@name='environment']/text()");

			Object result = expr.evaluate(doc, XPathConstants.STRING);
			if (result != null && result instanceof String) {
				try {
					env = EnvironmentEnum.valueOf(((String) result).toUpperCase());
				} catch (Throwable t) {
					cat.error("failed to convert environement name" + result, t);
				}
			}
		} catch (IOException e) {
			// file not exist or is not readable, ignore
		} catch (Throwable t) {
			cat.warn("failed to discover environment from host config file", t);
		}
		
		return env;
	}

	/**
	 * override a property at runtime
	 * @param path
	 * @param value
	 */
	public final synchronized void overrideProperty(String path, String value) {
		if (StringUtil.isNullOrEmpty(path)) throw new IllegalArgumentException("Property is null or empty");
		String[] tokens = path.split("/");
		Map currentMap = propsMap;
		for (int i = 0, len = tokens.length; i < len; i++) {
			if (i == len-1) {
				Object oldV = currentMap.get(tokens[i]);
				Object newV = null;
				if (oldV instanceof String) {
					newV = value;
				} else if (oldV instanceof Integer) {
					newV = new Integer(value);
				} else if (oldV instanceof Boolean) {
					newV = Boolean.valueOf(value);
				} else if (oldV instanceof Double) {
					newV = Double.valueOf(value);
				} else if (oldV instanceof List) {
					newV = Arrays.asList(value.split(","));
				} else if (oldV instanceof Map) {
					newV = new HashMap();
					String[] nvPair = value.split(",");
					for (int j = 0, len1 = nvPair.length; j < len1; j++) {
						String[] nv = nvPair[j].split("=");
						((HashMap) newV).put(nv[0], nv[1]);
					}
				} else if (oldV instanceof Class) {
					try {
						newV = Class.forName(value);
					} catch (ClassNotFoundException e) {
						cat.error("Failed to find class " + value + " for " + tokens[i]);
					}
				}
				if (newV != null) {
					currentMap.put(tokens[i], newV);
				}
			}
			else {
				currentMap = (Map) currentMap.get(tokens[i]);
			}
		}
		flatMap = null;
	}
	
	/**
	 * load configs from xml string passed in
	 * @param xmlString
	 */
	private synchronized void load(final String xmlString) 
	{
		try {
			DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
			domFactory.setNamespaceAware(true); // never forget this!
			DocumentBuilder builder = domFactory.newDocumentBuilder();

			// Read the entire document into memory
			// load default configs
			Document gProps = builder.parse(new StringInputStream(xmlString));
			getProperties(gProps);
		} catch(IOException e) {
			cat.error(e);
		} catch (SAXException e) {
			cat.error(e);
		} catch (ParserConfigurationException e) {
			cat.error(e);
		}

	}
	
	/**
	 * load all configs from xml files
	 * 
	 * @param path
	 * @param productName
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	private synchronized void load() {

		EnvironmentEnum environment = EnvironmentEnum.NONE;
		
		try {
			DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
			domFactory.setNamespaceAware(true); // never forget this!
			DocumentBuilder builder = domFactory.newDocumentBuilder();

			// Read the entire document into memory
			// load default configs
			Document gProps = builder.parse(new URL(path, product + ".props.xml").openStream());
			getProperties(gProps);

			// find out runtime environment
			if (RuntimeContext.getEnv() != null) {
				environment = EnvironmentEnum.valueOfStr(RuntimeContext.getEnv());
			}
			else {
				environment = deriveEnvironment();
			}

			// everything else is dev
			if (environment == EnvironmentEnum.NONE) {
				environment = EnvironmentEnum.DEV;
			}
			
			this.env = environment;

			// load environment specific configs
			try {
				Document eProps = builder.parse(new URL(path, product + "." + 
						environment.name() + ".props.xml").openStream());
				getProperties(eProps);
			} catch (IOException e) {
				cat.warn("Failed to read environment property xml file", e);
			} catch (SAXException e) {
				cat.warn("Failed to prase environment property xml file", e);
			}

			// load machine specific configs
			try {
				String host = InetAddress.getLocalHost().getHostName();
				String hostPropFilename = product + "."	+ host.toUpperCase() + ".props.xml";
				Document lProps = builder.parse(new URL(path, hostPropFilename).openStream());
				getProperties(lProps);
			} catch (IOException e) {
				cat.warn("Failed to read machine property xml file", e);
			} catch (SAXException e) {
				cat.warn("Failed to parse machine property xml file", e);
			}
			
		} catch (IOException e) {
			cat.fatal("Failed to read property xml file", e);
		} catch (ParserConfigurationException e) {
			cat.fatal("Failed to parse property xml file", e);
		} catch (SAXException e) {
			cat.fatal("Failed to parse property xml file", e);
		} finally {
			propsMap.put("environment/environment", environment.name());
		}
	}

	/**
	 * process xml document and populate properties
	 * 
	 * @param doc
	 */
	private void getProperties(Document doc) {
		NodeList children = doc.getDocumentElement().getChildNodes();
		for (int i = 0, len = children.getLength(); i < len; i++) {
			Node child = children.item(i);
			if (child.getNodeType() == Node.ELEMENT_NODE) {
				Map p = propsMap;
				getProperties(p, (Element) child);
			}
		}
	}

	/**
	 * process child element and populate properties
	 * 
	 * @param parentMap
	 * @param element
	 */
	static void getProperties(Map parentMap, Element element) {
		String nodeName = element.getNodeName();
		String attrName = element.getAttribute("name");
		if (NODE_SET.equalsIgnoreCase(nodeName)) {
			// a set
			Map childSet = null;
			if (parentMap.containsKey(attrName)) {
				childSet = (Map) parentMap.get(attrName);
			} else {
				childSet = new HashMap();
				parentMap.put(attrName, childSet);
			}
			NodeList children = element.getChildNodes();
			for (int i = 0, len = children.getLength(); i < len; i++) {
				Node child = children.item(i);
				if (child.getNodeType() == Node.ELEMENT_NODE) {
					getProperties(childSet, (Element) child);
				}
			}
		} else {
			// property
			String nodeValue = element.getTextContent();
			if (nodeValue != null) {
				nodeValue = nodeValue.trim();
			}
			Object value = null;
			if (!StringUtil.isNullOrEmpty(nodeValue)) {
				if (NODE_STRING.equalsIgnoreCase(nodeName)) {
					value = nodeValue;
				} else if (NODE_INT.equalsIgnoreCase(nodeName)) {
					value = new Integer(nodeValue);
				} else if (NODE_BOOLEAN.equalsIgnoreCase(nodeName)) {
					value = Boolean.valueOf(nodeValue);
				} else if (NODE_FLOAT.equalsIgnoreCase(nodeName)) {
					value = Double.valueOf(nodeValue);
				} else if (NODE_LIST.equalsIgnoreCase(nodeName)) {
					String[] values = nodeValue.split(",");
					if (values != null) {
						for (String v : values) {
							if (v!=null) v = v.trim();
						}
						value = Arrays.asList(values);
					} else {
						value = new ArrayList<String>();
					}
				} else if (NODE_MAP.equalsIgnoreCase(nodeName)) {
					value = new HashMap();
					String[] nvPair = nodeValue.split(",");
					for (int i = 0, len = nvPair.length; i < len; i++) {
						String[] nv = nvPair[i].split("=");
						((HashMap) value).put(nv[0], nv[1]);
					}
				}
			}
			parentMap.put(attrName, value);
		}
	}

	/**
	 * parse property set path
	 * 
	 * @param parent
	 * @param propSet
	 * @return
	 */
	static Map getPropSet(Map parent, String propSet) {
		int idx = propSet.indexOf('/');
		if (idx > 0) {
			String p = propSet.substring(0, idx);
			if (parent.containsKey(p)) {
				return getPropSet((Map) parent.get(p), propSet
						.substring(idx + 1));
			} else {
				throw new IllegalArgumentException(p + " set doesn't exist");
			}
		} else {
			if (parent.containsKey(propSet)) {
				return (Map) parent.get(propSet);
			} else {
				throw new IllegalArgumentException(propSet
						+ " set doesn't exist");
			}
		}
	}

	/**
	 * if a property exists
	 * 
	 * @param path
	 * @return
	 */
	public boolean exists(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return exists(path.substring(0, idx), path.substring(idx + 1));
	}

	/**
	 * get a boolean property
	 * 
	 * @param path
	 * @return
	 */
	public boolean getBoolean(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return getBoolean(path.substring(0, idx), path.substring(idx + 1),
				false);
	}

	/**
	 * get an int property
	 * 
	 * @param path
	 * @return
	 */
	public int getInt(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return getInt(path.substring(0, idx), path.substring(idx + 1), -1);
	}

	/**
	 * get a float property
	 * 
	 * @param path
	 * @return
	 */
	public double getFloat(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return getFloat(path.substring(0, idx), path.substring(idx + 1), 0.0);
	}

	/**
	 * get a string property
	 * 
	 * @param path
	 * @return
	 */
	public String getString(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return getString(path.substring(0, idx), path.substring(idx + 1), "");
	}

	/**
	 * get a list of string properties
	 * 
	 * @param path
	 * @return
	 */
	public List getList(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return getList(path.substring(0, idx), path.substring(idx + 1),
				new ArrayList());
	}

	/**
	 * get name/value pairs as a map
	 * 
	 * @param path
	 * @return
	 */
	public Map getMap(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return getMap(path.substring(0, idx), path.substring(idx + 1),
				new HashMap());
	}

	/**
	 * get class
	 * 
	 * @param path
	 * @return
	 */
	public Class getClass(String path) {
		int idx = path.lastIndexOf('/') > 0 ? path.lastIndexOf('/') : 0;
		return getClass(path.substring(0, idx), path.substring(idx + 1), null);
	}

	/**
	 * if a property exists
	 * 
	 * @param propSet
	 * @param propName
	 * @return
	 */
	public boolean exists(String propSet, String propName) {
		try {
			Map map = getPropSet(propsMap, propSet);
			if (map.containsKey(propName)) {
				return true;
			} else {
				return false;
			}
		} catch (IllegalArgumentException e) {
			return false;
		}
	}

	/**
	 * get a boolean property
	 * 
	 * @param propSet
	 * @param propName
	 * @param defaultValue
	 * @return
	 */
	public boolean getBoolean(String propSet, String propName,
			boolean defaultValue) {
		Map map = getPropSet(propsMap, propSet);
		if (map.containsKey(propName)) {
			if (map.get(propName) != null) {
				return ((Boolean) map.get(propName)).booleanValue();
			} else {
				return defaultValue;
			}
		} else {
			cat.warn("Boolean property " + propName
					+ " doesn't exist, use default value instead");
			return defaultValue;
		}
	}

	/**
	 * get an int property
	 * 
	 * @param propSet
	 * @param propName
	 * @param defaultValue
	 * @return
	 */
	public int getInt(String propSet, String propName, int defaultValue) {
		Map map = getPropSet(propsMap, propSet);
		if (map.containsKey(propName)) {
			if (map.get(propName) != null) {
				return ((Integer) map.get(propName)).intValue();
			} else {
				return defaultValue;
			}
		} else {
			cat.warn("Int property " + propName
					+ " doesn't exist, use default value instead");
			return defaultValue;
		}
	}

	/**
	 * get a float property
	 * 
	 * @param propSet
	 * @param propName
	 * @param defaultValue
	 * @return
	 */
	public double getFloat(String propSet, String propName, double defaultValue) {
		Map map = getPropSet(propsMap, propSet);
		if (map.get(propName) != null) {
			return ((Double) map.get(propName)).doubleValue();
		} else {
			cat.warn("Float property " + propName
					+ " doesn't exist, use default value " + defaultValue);
			return defaultValue;
		}
	}

	/**
	 * get a string property
	 * 
	 * @param propSet
	 * @param propName
	 * @param defaultValue
	 * @return
	 */
	public String getString(String propSet, String propName, String defaultValue) {
		Map map = getPropSet(propsMap, propSet);
		if (map.containsKey(propName)) {
			if (map.get(propName) != null) {
				return (String) map.get(propName);
			} else {
				return defaultValue;
			}
		} else {
			cat.warn("String property " + propName
					+ " doesn't exist, use default value instead");
			return defaultValue;
		}
	}

	/**
	 * get a list of string properties
	 * 
	 * @param propSet
	 * @param propName
	 * @param defaultList
	 * @return
	 */
	public ArrayList getList(String propSet, String propName,
			ArrayList defaultList) {
		Map map = getPropSet(propsMap, propSet);
		if (map.containsKey(propName)) {
			if (map.get(propName) != null) {
				return new ArrayList((List) map.get(propName));
			} else {
				return defaultList;
			}
		} else {
			cat.warn("List property " + propName
					+ " doesn't exist, use default list instead");
			return defaultList;
		}
	}

	/**
	 * get name/value pair as a map
	 * 
	 * @param propSet
	 * @param propName
	 * @param defaultMap
	 * @return
	 */
	public HashMap getMap(String propSet, String propName, HashMap defaultMap) {
		Map map = getPropSet(propsMap, propSet);
		if (map.containsKey(propName)) {
			if (map.get(propName) != null) {
				return (HashMap) map.get(propName);
			} else {
				return defaultMap;
			}
		} else {
			cat.warn("Map property " + propName
					+ " doesn't exist, use default map instead");
			return defaultMap;
		}
	}

	/**
	 * get class
	 * 
	 * @param propSet
	 * @param propName
	 * @param defaultClass
	 * @return
	 */
	public Class getClass(String propSet, String propName, Class defaultClass) {
		Map map = getPropSet(propsMap, propSet);
		if (map.containsKey(propName)) {
			if (map.get(propName) != null) {
				return (Class) map.get(propName);
			}
		}
		cat.warn("Class property " + propName
				+ " doesn't exist, use default class instead");
		return defaultClass;
	}

	/**
	 * get object
	 * @param propSet
	 * @param propName
	 * @param defValue
	 * @return
	 */
	public Object getObject(String propSet, String propName, Object defValue) {
		Map map = getPropSet(propsMap, propSet);
		if (map.containsKey(propName)) {
			if (map.get(propName) != null) {
				return map.get(propName);
			}
		}
		cat.warn("Class property " + propName + " doesn't exist, use default value instead");
		return defValue;
	}
	
	/**
	 * get all configs as a flat map
	 * 
	 * @return
	 */
	public synchronized HashMap getFlatMap() {
		if (flatMap != null)
			return flatMap;
		flatMap = new HashMap();
		convertToFlatMap(propsMap, flatMap, "");
		return flatMap;
	}

	/**
	 * convert all configs to a flat map
	 * 
	 * @param pm
	 * @param flatMap
	 * @param prefix
	 */
	public void convertToFlatMap(HashMap pm, HashMap flatMap, String prefix) {
		for (Iterator iter = pm.keySet().iterator(); iter.hasNext();) {
			String key = (String) iter.next();
			Object value = pm.get(key);
			if (value instanceof HashMap) {
				convertToFlatMap((HashMap) value, flatMap, prefix + key + '/');
			} else {
				flatMap.put(prefix + key, value);
			}
		}
	}

	/**
	 * render an xml to be used in component status page
	 * 
	 * @return
	 */
	public XMLElement asXml() {
		XMLElement configsElem = new XMLElement();
		configsElem.setName("Properties");
		configsElem.setAttribute("name", "Configurations");
		HashMap map = getFlatMap();
		List<String> keys = new ArrayList<String>();
		for (Object key : map.keySet()) {
			keys.add((String) key);
		}
		Collections.sort(keys);
		for (String configName : keys) {
			Object value = map.get(configName);
			int idx = configName.lastIndexOf('/');
			String configGroup = idx > 0 ? configName.substring(0,idx) : "";
			String configProp = idx > 0 ? configName.substring(idx+1) : configName;
			XMLElement configElem = new XMLElement();
			configElem.setName("Property");
			configElem.setAttribute("group", StringEscapeUtils.escapeXml(configGroup));
			configElem.setAttribute("name", StringEscapeUtils.escapeXml(configProp));
			configElem.setAttribute("class", (value!=null ? StringEscapeUtils.escapeXml(value.getClass().getName()) : ""));
			configElem.setContent((value!=null ? StringEscapeUtils.escapeXml(value.toString()) : ""));
			configsElem.addChild(configElem);
		}
		return configsElem;
	}



	/**
	 * render list of map in component status page
	 * @return
	 */
	public List getComponentStatusProperties() {
		List<Map<String, String>> props = new ArrayList<Map<String, String>>();
		HashMap configs = getFlatMap();
		List<String> configNames = null;
		List<String> t = new ArrayList<String>();
		for (Iterator iter = configs.keySet().iterator(); iter.hasNext();) {
			t.add((String) iter.next());
		}
		String[] st = (String[]) t.toArray(new String[0]);
		Arrays.sort(st);
		configNames = Arrays.asList(st);
		for (String configName : configNames) {
			Map<String, String> row = new HashMap<String, String>();
			Object value = configs.get(configName);
			int idx = configName.lastIndexOf('/');
			String configGroup = idx > 0 ? configName.substring(0,idx) : "";
			String configProp = idx > 0 ? configName.substring(idx+1) : configName;
			row.put("Group", configGroup);
			row.put("Name", configProp);
			row.put("Type", (value!=null ? value.getClass().getName() : ""));
			row.put("Value(s)", (value!=null ? value.toString() : ""));
			props.add(row);
		}
		return props;
	}
	
	public EnvironmentEnum getEnv() {
		return env;
	}

	public URL getPath() {
		return path;
	}

	public String getProduct() {
		return product;
	}
	
	static class DummyRuntimeConfigs implements IRuntimeConfigs {

		public void load(IConfigManager cfgMgr) {
			// noop
		}
		
	}

	static class Wrapper {
		ConfigManager mgr;
		IRuntimeConfigs configs;
		Wrapper(ConfigManager parser,
				IRuntimeConfigs configs) 
		{
			this.mgr = parser;
			this.configs = configs;
		}
	}

}
