package org.lightj.example.session;

import java.util.ArrayList;
import java.util.List;

import org.lightj.dal.SimpleLocatable;
import org.lightj.example.session.HelloWorldFlow.steps;
import org.lightj.example.task.SimpleASyncTask;
import org.lightj.example.task.SimpleASyncTaskMonitor;
import org.lightj.example.task.SimpleASyncTaskRunner;
import org.lightj.session.FlowExecutionException;
import org.lightj.session.FlowResult;
import org.lightj.session.FlowSession;
import org.lightj.session.FlowSessionFactory;
import org.lightj.session.FlowState;
import org.lightj.session.step.AsyncStepBuilder;
import org.lightj.session.step.StepBuilder;
import org.lightj.session.step.DelayedEnclosure;
import org.lightj.session.step.FlowStepBuilder;
import org.lightj.session.step.IFlowStep;
import org.lightj.session.step.MappedCallbackHandler;
import org.lightj.session.step.MappedErrorHandler;
import org.lightj.session.step.RetryEnclosure;
import org.lightj.session.step.SimpleStepExecution;
import org.lightj.session.step.StepErrorHandler;
import org.lightj.session.step.StepExecution;
import org.lightj.session.step.StepTransition;
import org.lightj.task.MonitorOption;
import org.lightj.task.Task;
import org.lightj.task.TaskMonitor;
import org.lightj.task.TaskResult;
import org.lightj.task.TaskResultEnum;

/**
 * a spring factory creates steps in a workflow
 * 
 * @author binyu
 *
 */
@SuppressWarnings("rawtypes")
public class HelloWorldFlowStepsImpl {
	
	/** shared error handler */
	private static StepErrorHandler<HelloWorldFlowContext> errorHandler = new MappedErrorHandler<HelloWorldFlowContext>(new StepTransition(steps.error));

	/**
	 * a simple async task
	 * @return
	 */
	public static IFlowStep buildAsyncTaskStep() {
		
		Task task = new SimpleASyncTask(2);
		MonitorOption monitorOption = MonitorOption.newInstance(0, 1000, 0);
		TaskMonitor<? extends Task> taskMonitor = new SimpleASyncTaskMonitor(monitorOption);
		
		// result handler, with default transition
		MappedCallbackHandler<HelloWorldFlowContext> resultHandler = new MappedCallbackHandler<HelloWorldFlowContext>(
				new StepTransition().toStep(steps.error).withResult(FlowResult.Failed)
		);
		// success result
		resultHandler.mapResultTo(TaskResultEnum.Success, new StepTransition().toStep(steps.sessionJoinStep));

		// build the step
		return new AsyncStepBuilder().executeAsyncPollTask(new SimpleASyncTaskRunner(), 
				taskMonitor, 
				errorHandler, 
				resultHandler, 
				errorHandler, 
				task).getFlowStep();
	}
	
	/**
	 * build join step
	 * @return
	 */
	public static IFlowStep buildJoinStep() 
	{
		List<FlowSession> children = new ArrayList<FlowSession>();
		for (int i = 0; i < 2; i++) {
			FlowSession child = FlowSessionFactory.getInstance().createSession(
					HelloWorldFlow.class, 
					new SimpleLocatable(),
					new SimpleLocatable());
			children.add(child);
		}

		// result handler, with default transition
		MappedCallbackHandler<HelloWorldFlowContext> resultHandler = new MappedCallbackHandler<HelloWorldFlowContext>(
				new StepTransition().toStep(steps.error).withResult(FlowResult.Failed)
		) {

			@Override
			public synchronized StepTransition executeOnResult(TaskResult result) throws FlowExecutionException {
				// remember result
				sessionContext.setSplitCount(sessionContext.getSplitCount() + 1);
				return super.executeOnResult(result);
			}
			
		};

		// success result
		resultHandler.mapResultTo(TaskResultEnum.Success, new StepTransition().toStep(steps.delayStep));

		// build the step
		return new FlowStepBuilder<HelloWorldFlowContext>().launchSubFlows(children, errorHandler, resultHandler, errorHandler).getFlowStep();
	}

	/**
	 * build a step with delay
	 * @return
	 */
	public static IFlowStep buildDelayStep() 
	{
		StepExecution execution = new SimpleStepExecution<HelloWorldFlowContext>(new StepTransition().toStep(steps.retryStep));
		IFlowStep step = new StepBuilder().execute(execution, errorHandler).getFlowStep();
		return DelayedEnclosure.delay(3000, step);		
	}

	/**
	 * build a retry step
	 * @return
	 */
	public static IFlowStep buildRetryStep() 
	{
		StepExecution execution = new SimpleStepExecution<HelloWorldFlowContext>(new StepTransition().toStep(steps.delayWithOptionStep)) {
			@Override
			public StepTransition execute() throws FlowExecutionException {
				this.sessionContext.setRunCount(this.sessionContext.getRunCount() + 1);
				return super.execute();
			}
		};
		IFlowStep step = new StepBuilder().execute(execution, errorHandler).getFlowStep();
		StepTransition matchTran = StepTransition.runToStep(steps.delayWithOptionStep);
		return RetryEnclosure.retry(matchTran, 3, 1000, step);		
	}
	
	/**
	 * build a step with delay with step option
	 * @return
	 */
	public static IFlowStep buildDelayStepWithStepOption() 
	{
		StepExecution execution = new SimpleStepExecution<HelloWorldFlowContext>(new StepTransition().toStep(steps.timeoutStep));
		IFlowStep step = new StepBuilder().execute(execution, errorHandler).getFlowStep();
		step.setDelayMilliSec(3000);
		return step;
	}
	

	/**
	 * build a step with timeout option
	 * @return
	 */
	public static IFlowStep buildTimeoutStep() 
	{
		StepExecution execution = new SimpleStepExecution<HelloWorldFlowContext>(new StepTransition().inState(FlowState.Callback)) {
			@Override
			public StepTransition execute() throws FlowExecutionException {
				return defResult;
			}
			
		}; 
		
		StepErrorHandler<HelloWorldFlowContext> timeoutHandler = new MappedErrorHandler<HelloWorldFlowContext>(new StepTransition(steps.stop)) {
			
			@Override
			public StepTransition executeOnError(Throwable t) {
				assert t instanceof FlowExecutionException;
				this.sessionContext.setTimeoutCount(this.sessionContext.getTimeoutCount()+1);
				return defResult;
			}
		};
		
		
		IFlowStep step = new StepBuilder().execute(execution, timeoutHandler).getFlowStep();
		step.setTimeoutMilliSec(2000);
		return step;
	}
}
