package org.lightj.jms;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.TextMessage;

import org.lightj.concurrent.CircularArray;
import org.lightj.concurrent.CounterStats;
import org.lightj.config.LightJRuntimeConfigs;
import org.lightj.util.StringUtil;
import org.lightj.util.nanoxml.XMLElement;


/**
 * collect group stats
 * 
 * @author biyu
 *
 */
public final class ConnectionStats extends CounterStats {
	
	static SimpleDateFormat fmt = new SimpleDateFormat("yyyy-MM-dd kk:mm:ss.SSS");

	/** worst execution duration in ms */
	private volatile long worstExecutionDuration = 0L;
	
	/** resetable execution count */
	private volatile int producedTotal = 0;
	private volatile int consumedTotal = 0;
	
	/** error msg */
	private String errorCode;
	
	/** error time */
	private Date errorTime;
	
	/** connection and session info */
	private String providerUrl;
	private String sessName;
	private boolean producing;
	private boolean consuming;
	
	private ExecutorService producingExecutor;
	
	/** message history */
	private CircularArray<String> msgProduceHistory;
	private CircularArray<String> msgConsumeHistory;
	
	public ConnectionStats() {
		super("", "JmsConnectionStats");
	}
	
	/** constructor */
	public ConnectionStats(String providerUrl, String sessName, boolean producing, boolean consuming) {
		super(providerUrl, "JmsConnectionStats");
		this.providerUrl = providerUrl;
		this.sessName = sessName;
		this.producing = producing;
		this.consuming = consuming;
		this.msgProduceHistory = new CircularArray<String>(new String[LightJRuntimeConfigs.jmsHistorySize]);
		this.msgConsumeHistory = new CircularArray<String>(new String[LightJRuntimeConfigs.jmsHistorySize]);
	}
	
	/**
	 * recalculate execution durations (worst and mean)
	 * @param executionDuration
	 */
	public void adjustExecutionDuration(long executionDuration) {
		worstExecutionDuration = Math.max(worstExecutionDuration, executionDuration);
	}
	
	void setProducingExecutor(ExecutorService producingExecutor) {
		this.producingExecutor = producingExecutor;
	}

	/** reset all stats */
	public synchronized void reset() {
		super.reset();
		worstExecutionDuration = 0;
	}
	
	/** a message is going out */
	public void onOutgoingMessage(OutboundMessage outMsg) {
		producedTotal++;
		incTotalCount();
		msgProduceHistory.replace(outMsg.toString());
	}
	
	/** a message is coming in */
	public void onIncomingMessage(Message inMsg) {
		consumedTotal++;
		incTotalCount();
		StringBuffer buf = new StringBuffer();
		if (inMsg instanceof TextMessage) {
			try {
				buf.append("Time=").append(fmt.format(new Date())).append(',');
				buf.append("CorrelationID=").append(inMsg.getJMSCorrelationID()).append(',');
				buf.append("Content=").append(StringUtil.trimToLength(((TextMessage) inMsg).getText(), 40));
			} catch (Throwable t) {
				buf.append("Error:").append(t.getMessage());
			}
		}
		else {
			buf.append("Received a non-text message");
		}
		msgConsumeHistory.replace(buf.toString());
	}
	
	/**
	 * exception happens
	 * @param e
	 */
	public void onException(JMSException e) {
		errorCode = e.getErrorCode();
		errorTime = new Date();
	}		
	
	/** reset the size of the history */
	public synchronized void resetSize(int newSize) {
		int size = Math.max(msgProduceHistory.size(), newSize);
		for (int i = msgProduceHistory.size(); i < size; i++) {
			msgProduceHistory.add(null);
		}
		size = Math.max(msgConsumeHistory.size(), newSize);
		for (int i = msgConsumeHistory.size(); i < size; i++) {
			msgConsumeHistory.add(null);
		}
	}

	@Override
	protected void asXml(XMLElement statElem) {
		super.asXml(statElem);
		statElem.setAttribute("providerURL", providerUrl);
		statElem.setAttribute("sessionName", sessName);
		statElem.setAttribute("producing", Boolean.toString(producing));
		if (producing) {
			statElem.setAttribute("totalProduced", producedTotal);
			if (producingExecutor != null && producingExecutor instanceof ThreadPoolExecutor) {
				statElem.setAttribute("producingQueueSize", Integer.toString(((ThreadPoolExecutor)producingExecutor).getQueue().size()));
			}
		}
		statElem.setAttribute("consuming", Boolean.toString(consuming));
		if (consuming) {
			statElem.setAttribute("totalConsumed", consumedTotal);
		}
		statElem.setAttribute("worstExecutionDuration", TimeUnit.MILLISECONDS.convert(worstExecutionDuration, TimeUnit.NANOSECONDS));
		statElem.setAttribute("lastError", errorTime!=null ? errorTime.toString() : "n/a");
		statElem.setAttribute("errorCode", errorCode!=null ? errorCode : "n/a");

		// histories
		if (producing) {
			XMLElement produceElem = new XMLElement();
			produceElem.setName("ProducingHistory");
			statElem.addChild(produceElem);
			for (int i = 0, size = msgProduceHistory.size(); i < size; i++) {
				String msg = msgProduceHistory.next();
				if (!StringUtil.isNullOrEmpty(msg)) {
					XMLElement msgElem = new XMLElement();
					msgElem.setName("JmsMsg");
					msgElem.setAttribute("content", msg);
					produceElem.addChild(msgElem);
				}
			}
		}
		if (consuming) {
			XMLElement consumeElem = new XMLElement();
			consumeElem.setName("ConsumingHistory");
			statElem.addChild(consumeElem);
			for (int i = 0, size = msgConsumeHistory.size(); i < size; i++) {
				String msg = msgConsumeHistory.next();
				if (!StringUtil.isNullOrEmpty(msg)) {
					XMLElement msgElem = new XMLElement();
					msgElem.setName("JmsMsg");
					msgElem.setAttribute("Content", msg);
					consumeElem.addChild(msgElem);
				}
			}
		}
	}

}
