package org.lightj.jms;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.jms.DeliveryMode;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueReceiver;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.jms.TopicSubscriber;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.lightj.concurrent.PriorityRoundRobinThreadPoolExecutor;
import org.lightj.config.LightJRuntimeConfigs;
import org.lightj.util.Log4jProxy;
import org.lightj.util.NetUtil;
import org.lightj.util.StringUtil;


/**
 * manager for JMS messages
 * @author biyu
 *
 */
public class MessageManager {
	
	/** logger */
	static Log4jProxy cat = Log4jProxy.getInstance(MessageManager.class.getName());
	
	/** host name, used as part of unique msgid */
	static String hostName = NetUtil.getMyHostName();
	
	/** singleton */
	private static final MessageManager singleton = new MessageManager();
	
	/** singleton */
	public static MessageManager getInstance() {
		return singleton;
	}
	
	/** generate a unique msg id */
	public static String genMessageId() {
		return NetUtil.getMyHostName() + '-' + Long.toString(Thread.currentThread().getId()) + '-' + Long.toString(System.nanoTime());
	}
	
	/** a collection of all existing sessions, used in reconnect all */
	private static Set<JmsSessionWrapper> existingSess = new HashSet<JmsSessionWrapper>();
	
	/** a map of providerUrl to {@link TopicConnection} */
	private Map<String, TopicConnection> topicConns = new HashMap<String, TopicConnection>();
	/** a map of providerUrl to {@link QueueConnection} */
	private Map<String, QueueConnection> queueConns = new HashMap<String, QueueConnection>();
	
	/** a map of topic name to {@link TopicConnection} */
	private Map<String, TopicConnection> topic2Conns = new HashMap<String, TopicConnection>();
	/** a map of queue name to {@link QueueConnection} */
	private Map<String, QueueConnection> queue2Conns = new HashMap<String, QueueConnection>();
	
	/** a map of topic name to outbound publisher executor service */
	private Map<String, ExecutorService> topicExecutors = new HashMap<String, ExecutorService>();
	/** a map of queue name to outbound sender executor service */
	private Map<String, ExecutorService> queueExecutors = new HashMap<String, ExecutorService>();
	
	/** default msg expiration time, default is never */
	private long defExpiration = Message.DEFAULT_TIME_TO_LIVE;
	
	/**
	 * Constructor 
	 */
	private MessageManager() {
		defExpiration = LightJRuntimeConfigs.messageTTL;
	}
	
	public synchronized void setup(JmsConfig jmsCfg) throws JMSException, NamingException {
		if (jmsCfg.isQueue()) {
			setupQueue(jmsCfg);
		}
		else {
			setupTopic(jmsCfg);
		}
	}
	
	/**
	 * This method initialises a topic, 
	 * 1. cache or setup connection
	 * 2. create threadpool of size 1 to publish (with its own session)
	 * 3. attach msg listener (with its own session)
	 * 4. attach ex handler if any
	 * @throws NamingException 
	 * @throws Exception
	 */
	public synchronized void setupTopic(JmsConfig sessProps) throws JMSException, NamingException 
	{
		JmsSessionWrapper sessionWrapper = new JmsSessionWrapper(sessProps);
		TopicConnection conn = null;
		if (topicConns.containsKey(sessProps.getProviderUrl())) {
			// we have an existing connection to this broker
			conn = topicConns.get(sessProps.getProviderUrl());
		}
		else {
			// create a new connection to this broker and cache it
			Properties props = new Properties();
			props.setProperty(Context.INITIAL_CONTEXT_FACTORY, LightJRuntimeConfigs.jmsIntialContextFactory);
			props.setProperty(Context.PROVIDER_URL, sessProps.getProviderUrl());
			Context ctx = new InitialContext(props);

			TopicConnectionFactory factory = (TopicConnectionFactory) ctx.lookup(LightJRuntimeConfigs.jmsConnectionFactory);
			cat.info("jms connection factory found");
			
			if(!StringUtil.isNullOrEmptyAfterTrim(sessProps.getUsername()) &&
					!StringUtil.isNullOrEmptyAfterTrim(sessProps.getPassword()))
			{
				conn = factory.createTopicConnection(sessProps.getUsername(), sessProps.getPassword());
			} 
			else {
				conn = factory.createTopicConnection();
			}
			// attach ExceptionListener wrapper to this connection
			conn.setExceptionListener(new ExceptionListenerWrapper(sessionWrapper));
			// cache it so it can be reused for others with the same connection
			topicConns.put(sessProps.getProviderUrl(), conn);
			cat.info("conection created");
		}
		
		topic2Conns.put(sessProps.getName(), conn);
		if (sessProps.getExHandler() != null) {
			((ExceptionListenerWrapper) conn.getExceptionListener()).addListener(sessProps.getExHandler());
			cat.info("jms exception handler added");
		}
		
		try {
			if (sessProps.isConsuming() && sessProps.getMsgListener() != null) {
				// create a session for each subscriber
				TopicSession session = conn.createTopicSession(false, sessProps.getConsumingAck());
				cat.info("jms session is created");
				Topic topic = session.createTopic(sessProps.getName());
				cat.info("jms topic is created");
				TopicSubscriber subscriber = null;
				if (sessProps.isDurable()) {
					if (sessProps.getMsgSelector() != null) {
						subscriber = session.createDurableSubscriber(topic, sessProps.getName(), sessProps.getMsgSelector(), sessProps.isNoLocal());
					}
					else {
						subscriber = session.createDurableSubscriber(topic, sessProps.getName() + '_' + NetUtil.getMyHostName());
					}
				}
				else {
					if (sessProps.getMsgSelector() != null) {
						subscriber = session.createSubscriber(topic, sessProps.getMsgSelector(), sessProps.isNoLocal());
					}
					else {
						subscriber = session.createSubscriber(topic);
					}					
				}
				cat.info("jms subscriber is created");
				// create MessageListener wrapper
				boolean useInternalExecutor = (sessProps.getConsumerExecutor()==null);
				ExecutorService executor = !useInternalExecutor ? sessProps.getConsumerExecutor() : Executors.newSingleThreadExecutor();
				MessageListenerWrapper listener = new MessageListenerWrapper(sessionWrapper, sessProps.getMsgListener(), executor);
				subscriber.setMessageListener(listener);
				sessionWrapper.addConsumer(session, subscriber, executor, useInternalExecutor);
				cat.info("jms listeners added");
			}
			if (sessProps.isProducing()) {
				if (topicExecutors.containsKey(sessProps.getName())) {
					// one queue per topic
					cat.info("outbound queue already created for topic : " + sessProps.getName());
				}
				else {
					// create a queue and attach # of threads as publisher
					ExecutorService executor = Executors.newFixedThreadPool(Math.max(1, sessProps.getNumOfProducingThread()));
					topicExecutors.put(sessProps.getName(), executor);
					sessionWrapper.setProducerExecutor(executor);
				}
			}
		} catch (JMSException e) {
			cat.error("failed jms with jms exception", e);
			throw e;
		}
		// if we reach here, save the original session info for later reconnect on demand
		existingSess.add(sessionWrapper);
	}
	
	/**
	 * This method initialises a queue, 
	 * 1. cache or setup connection
	 * 2. create threadpool of size 1 to send (with its own session)
	 * 3. attach msg listener (with its own session)
	 * 4. attach ex handler if any
	 * 
	 * @param connProps
	 * @param sessProps
	 * @throws JMSException
	 * @throws NamingException
	 */
	public synchronized void setupQueue(JmsConfig sessProps) 
		throws JMSException, NamingException
	{
		JmsSessionWrapper sessionWrapper = new JmsSessionWrapper(sessProps);
		QueueConnection conn = null;
		if (queueConns.containsKey(sessProps.getProviderUrl())) {
			// we have an existing connection to this broker
			conn = queueConns.get(sessProps.getProviderUrl());
		}
		else {
			// create a new connection to this broker and cache it
			Properties props = new Properties();
			props.setProperty(Context.INITIAL_CONTEXT_FACTORY, LightJRuntimeConfigs.jmsIntialContextFactory);
			props.setProperty(Context.PROVIDER_URL, sessProps.getProviderUrl());
			Context ctx = new InitialContext(props);

			QueueConnectionFactory factory = (QueueConnectionFactory) ctx.lookup(LightJRuntimeConfigs.jmsConnectionFactory);
			cat.info("jms connection factory found");
			
			if(!StringUtil.isNullOrEmptyAfterTrim(sessProps.getUsername())
					&& !StringUtil.isNullOrEmptyAfterTrim(sessProps.getPassword()))
			{
				conn = factory.createQueueConnection(sessProps.getUsername(), sessProps.getPassword());
			}
			else {
				conn = factory.createQueueConnection();
			}
			conn.setExceptionListener(new ExceptionListenerWrapper(sessionWrapper));
			queueConns.put(sessProps.getProviderUrl(), conn);
			cat.info("conection created");
		}
		queue2Conns.put(sessProps.getName(), conn);
		if (sessProps.getExHandler() != null) {
			((ExceptionListenerWrapper) conn.getExceptionListener()).addListener(sessProps.getExHandler());
			cat.info("jms exception handler added");
		}
		
		try {
			if (sessProps.isConsuming() && sessProps.getMsgListener() != null) {
				// create a session for each subscriber
				QueueSession session = conn.createQueueSession(false, sessProps.getConsumingAck());
				cat.info("jms session is created");
				Queue queue = session.createQueue(sessProps.getName());
				cat.info("jms queue is created");
				QueueReceiver receiver = null;
				// queue is always durable
				if (sessProps.getMsgSelector() != null) {
					receiver = session.createReceiver(queue, sessProps.getMsgSelector());
				}
				else {
					receiver = session.createReceiver(queue);
				}					
				cat.info("jms receiver is created");
				// create MessageListener wrapper
				boolean useInternalExecutor = (sessProps.getConsumerExecutor() == null);
				ExecutorService executor = !useInternalExecutor ? sessProps.getConsumerExecutor() : Executors.newSingleThreadExecutor();
				MessageListenerWrapper listener =  new MessageListenerWrapper(sessionWrapper, sessProps.getMsgListener(), executor);
				receiver.setMessageListener(listener);
				sessionWrapper.addConsumer(session, receiver, executor, useInternalExecutor);
				cat.info("jms listeners added");
			}
			if (sessProps.isProducing()) {
				if (queueExecutors.containsKey(sessProps.getName())) {
					// one queue per topic
					cat.info("outbound queue already created for queue : " + sessProps.getName());
				}
				else {
					// create a queue and attach # of threads as publisher
					ExecutorService executor = Executors.newFixedThreadPool(Math.max(1, sessProps.getNumOfProducingThread()));
					queueExecutors.put(sessProps.getName(), executor);
					sessionWrapper.setProducerExecutor(executor);
				}
			}
		} catch (JMSException e) {
			cat.error("failed jms with jms exception", e);
			throw e;
		}
		// if we reach here, save the original session info for later reconnect on demand
		existingSess.add(sessionWrapper);
	}

	/**
	 * start a connection for a provider(broker) url
	 * @param providerUrl
	 */
	public synchronized void start(String providerUrl) throws JMSException {
		if (topicConns.containsKey(providerUrl)) {
			topicConns.get(providerUrl).start();
		}
		if (queueConns.containsKey(providerUrl)) {
			queueConns.get(providerUrl).start();
		}
	}

	/**
	 * start all avaialble connections
	 * @throws JMSException 
	 *
	 */
	public synchronized void startAll() throws JMSException {
		StringBuffer buf = new StringBuffer();
		for (String url : topicConns.keySet()) {
			TopicConnection conn = topicConns.get(url);
			try {
				conn.start();
			} catch (JMSException e) {
				cat.error("topic failed to start", e);
				buf.append(url).append(" failed to start\n");
			}
		}
		for (String url : queueConns.keySet()) {
			QueueConnection conn = queueConns.get(url);
			try {
				conn.start();
			} catch (JMSException e) {
				cat.error("queue failed to start", e);
				buf.append(url).append(" failed to start\n");
			}
		}
		if (buf.length() > 0) {
			throw new JMSException(buf.toString());
		}
	}

	/**
	 * shutdown a {@link MessageListener} an underneath jms consumer/session
	 * @param providerUrl
	 * @param destName
	 * @param listenerKlass
	 * @throws JMSException
	 */
	public final synchronized void shutdownConsumer(JmsConfig sp) throws JMSException {
		for (JmsSessionWrapper w : existingSess) {
			JmsConfig cfg = w.sessProp;
			if (StringUtil.equalIgnoreCase(cfg.getProviderUrl(), sp.getProviderUrl()) &&
					StringUtil.equalIgnoreCase(cfg.getName(), sp.getName()) &&
					cfg.isConsuming() &&
					cfg.getMsgListener().getClass() == sp.getMsgListener().getClass()) {
				w.jmsConsumer.shutdown();
			}
		}		
	}
	
	/**
	 * Method to cleanup the TopicConnection during shutdown 
	 * 
	 * @return TopicConnection
	 */
	public final synchronized void shutdown() {
		// stop all executors first so we don't try to produce any message while connection is being shutdown
		for (ExecutorService queueExecutor : queueExecutors.values()) {
			final ExecutorService pool = queueExecutor;
			new Thread(new Runnable() {
				public void run() {
					pool.shutdown(); // Disable new tasks from being submitted
					try {
						// Wait a while for existing tasks to terminate
						if (!pool.awaitTermination(10, TimeUnit.SECONDS)) {
							pool.shutdownNow();
						}
					} catch (InterruptedException ie) {
						pool.shutdownNow();
						Thread.currentThread().interrupt();
					}
				}
			}).start();
		}

		for (ExecutorService topicExecutor : topicExecutors.values()) {
			final ExecutorService pool = topicExecutor;
			new Thread(new Runnable() {
				public void run() {
					pool.shutdown(); // Disable new tasks from being submitted
					try {
						// Wait a while for existing tasks to terminate
						if (!pool.awaitTermination(10, TimeUnit.SECONDS)) {
							pool.shutdownNow();
						}
					} catch (InterruptedException ie) {
						pool.shutdownNow();
						Thread.currentThread().interrupt();
					}
				}
			}).start();
		}

		// now stop the jms connections
		try {
			for (TopicConnection conn : topicConns.values()) {
				try {
					conn.stop();
					conn.close();
					conn = null;
				} catch (JMSException e) {
					cat.error("failed to shutdown jms connection", e);
				}
			}
		} catch (Throwable t) {
			cat.error(t);
		} finally {
			topicConns.clear();
			topic2Conns.clear();
			topicExecutors.clear();
		}
		try {
			for (QueueConnection conn : queueConns.values()) {
				try {
					conn.stop();
					conn.close();
					conn = null;
				} catch (JMSException e) {
					cat.error("failed to shutdown jms connection", e);
				}
			}
		} catch (Throwable t) {
			cat.error(t);
		} finally {
			queueConns.clear();
			queue2Conns.clear();
			queueExecutors.clear();
		}
	}
	
	/**
	 * refresh all connections
	 *
	 */
	synchronized final void restartAll() {
		// shutdown existing connections
		shutdown();
		
		// reconnect
		final Set<JmsSessionWrapper> tobeRestarted = new HashSet<JmsSessionWrapper>(existingSess);
		existingSess.clear();
		for (JmsSessionWrapper sess : tobeRestarted) {
			try {
				setup(sess.sessProp);
			} catch (NamingException e) {
				cat.error("Restart failed for " + sess.sessProp.getName(), e);
			} catch (JMSException e) {
				cat.error("Restart failed for " + sess.sessProp.getName(), e);
			}
		}
		
		// start all connections
		try {
			startAll();
		} catch (JMSException e) {
			cat.error("Restart connections failed ", e);
		}
	}

	/**
	 * set jms default expiration time
	 * @param expiration
	 */
	public void setJmsDefaultExpiration(long expiration) {
		this.defExpiration = expiration;
	}
	
	/**
	 * produce message to topic or queue
	 * @param destination
	 * @param msg
	 * @return
	 * @throws Exception
	 */
	public OutboundMessage produce(String destination, String msg) throws Exception {
		return produce(destination, msg, null);		
	}
	
	/**
	 * produce message to topic or queue
	 * @param destination
	 * @param msg
	 * @return
	 * @throws Exception
	 */
	public OutboundMessage produce(String destination, String msg, ExceptionListener handler) throws Exception {
		OutboundMessage message = OutboundMessageImpl.newInstance(
				null, 
				Session.AUTO_ACKNOWLEDGE, 
				defExpiration, 
				Message.DEFAULT_PRIORITY, 
				System.currentTimeMillis(), 
				genMessageId(), 
				msg);
		return produce(destination, message, handler);		
	}

	/**
	 * produce message to topic or queue
	 * @param destination
	 * @param message
	 * @return
	 * @throws Exception
	 */
	public OutboundMessage produce(String destination, OutboundMessage message) throws Exception {
		return produce(destination, message, null);
	}
	
	/**
	 * produce message to topic or queue, attach an exception listener to get callback when exception happens
	 * @param destination
	 * @param message
	 * @return
	 * @throws Exception
	 */
	public OutboundMessage produce(String destination, OutboundMessage message, ExceptionListener handler) throws Exception {
		if (topicExecutors.containsKey(destination)) {
			topicExecutors.get(destination).submit(new TopicWorker(destination, message, handler));
			return message;
		} else if (queueExecutors.containsKey(destination)) {
			queueExecutors.get(destination).submit(new QueueWorker(destination, message, handler));
			return message;
		}
		throw new Exception("JMS destination " + destination + " doesn't exist or cannot be published to");
	}
	

	/**
	 * publish a message to an existing topic
	 * @param topicName
	 * @param msg
	 * @throws Exception
	 */
	public OutboundMessage publish2Topic(String topicName, String msg) throws Exception {
		OutboundMessage message = OutboundMessageImpl.newInstance(
				null, 
				Session.AUTO_ACKNOWLEDGE, 
				defExpiration, 
				Message.DEFAULT_PRIORITY, 
				System.currentTimeMillis(), 
				genMessageId(), 
				msg);
		return publish2Topic(topicName, message);
	}
	
	
	/**
	 * This method publish message on Topic
	 * @param topicName
	 * @param msg
	 * @throws Exception
	 */
	public OutboundMessage publish2Topic(String topicName, OutboundMessage message) throws Exception {
		return produce(topicName, message);
	}
	
	/**
	 * send a message to an existing queue
	 * @param queueName
	 * @param msg
	 * @throws Exception
	 */
	public OutboundMessage send2Queue(String queueName, String msg) throws Exception {
		OutboundMessage message = OutboundMessageImpl.newInstance(
				null, 
				Session.AUTO_ACKNOWLEDGE, 
				defExpiration, 
				Message.DEFAULT_PRIORITY, 
				System.currentTimeMillis(), 
				genMessageId(), 
				msg);
		return send2Queue(queueName, message);
	}
	
	/**
	 * send a message to an existing queue
	 * @param queueName
	 * @param message
	 * @return
	 * @throws Exception
	 */
	public OutboundMessage send2Queue(String queueName, OutboundMessage message) throws Exception {
		return produce(queueName, message);
	}
	
	/**
	 * create a topic publisher, 
	 * no need to be synchronized since this method is always executed exactly once within each thread's context 
	 * @param topicName
	 * @return
	 * @throws JMSException 
	 */
	static TopicWrapper createTopicPublisher(String topicName, long expiration, int deliveryMode) throws JMSException {
		try {
			TopicConnection conn = getInstance().topic2Conns.get(topicName);
			TopicSession session = conn.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
			Topic topic = session.createTopic(topicName);
			TopicPublisher publisher = session.createPublisher(topic);
			publisher.setTimeToLive(expiration);
			publisher.setDeliveryMode(deliveryMode);
			cat.info("jms publisher is created");
			// now find the stats and attache to wrapper so we can collect stats later
			ConnectionStats stats = new ConnectionStats();
			for (JmsSessionWrapper sw : existingSess) {
				JmsConfig cfg = sw.sessProp;
				if (StringUtil.equalIgnoreCase(cfg.getName(), topicName) 
						&& cfg.isProducing() && !cfg.isQueue()) {
					stats = sw.getStats();
					break;
				}
			}
			return new TopicWrapper(session, publisher, stats);
		} catch (JMSException e) {
			cat.error(e);
			throw e;
		}
	}
	
	/**
	 * topic wrapper
	 * @author biyu
	 *
	 */
	static class TopicWrapper {
		/** JMS topic session created */
		TopicSession sess;
		/** JMS publisher created */
		TopicPublisher pub;
		/** stats collector */
		ConnectionStats stats;
		/** constructor */
		TopicWrapper(TopicSession sess, TopicPublisher pub, ConnectionStats stats) {
			this.sess = sess;
			this.pub = pub;
			this.stats = stats;
		}
	}

	/**
	 * a working {@link Runnable} that does the publish work
	 * @author biyu
	 *
	 */
	static class TopicWorker implements Runnable {
		
		private static ThreadLocal<TopicWrapper> publishers = new ThreadLocal<TopicWrapper>() {};

		private String topicName;
		private OutboundMessage msg;
		private ExceptionListener handler;
		
		TopicWorker(String topicName, OutboundMessage msg) {
			this(topicName, msg, null);
		}
		
		TopicWorker(String topicName, OutboundMessage msg, ExceptionListener handler) {
			this.topicName = topicName;
			this.msg = msg;
			this.handler = handler;
		}
		
		public void run() {
			TopicWrapper publisher = publishers.get();
			try {
				if (publisher == null) {
					publisher = createTopicPublisher(topicName, Message.DEFAULT_TIME_TO_LIVE, DeliveryMode.NON_PERSISTENT);
					publishers.set(publisher);
				}
				TextMessage tmsg = publisher.sess.createTextMessage(msg.getM_JMSContent());
				tmsg.setJMSCorrelationID(msg.getM_JMSCorrelationID());
				tmsg.setJMSDeliveryMode(msg.getM_JMSDeliveryMode());
				tmsg.setJMSPriority(msg.getM_JMSPriority());
				tmsg.setJMSExpiration(msg.getM_JMSExpiration());
				tmsg.setJMSTimestamp(msg.getM_JMSTimestamp());
				Map<String, Object> customHeaders = msg.getReadOnlyJMSProperties();
				if(customHeaders != null){
					for (Iterator<String> iter = customHeaders.keySet().iterator(); iter.hasNext(); ) {
						String key = (String)iter.next();
						Object value = customHeaders.get(key);
						tmsg.setObjectProperty(key, value);
					}
				}
				publisher.pub.publish(tmsg);
				publisher.stats.onOutgoingMessage(msg);
				customHeaders = null;
			} catch (final JMSException e) {
				cat.error(e);
				if (handler != null) {
					// we don't want to block the publisher
					PriorityRoundRobinThreadPoolExecutor.getInstance().execute(
							new JmsQueueTask (new Runnable() {
								public void run() {
									handler.onException(new AsyncJmsException(e, msg));
								}
							})
					);
				}
			}
		}
	}
	
	/**
	 * create a queue sender
	 * no need to be synchronized since this method is always executed exactly once within each thread's context 
	 * @param queueName
	 * @return
	 * @throws JMSException 
	 */
	static QueueWrapper createQueueSender(String queueName) throws JMSException {
		try {
			QueueConnection conn = getInstance().queue2Conns.get(queueName);
			QueueSession session = conn.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);
			Queue queue = session.createQueue(queueName);
			QueueSender sender = session.createSender(queue);
			cat.info("jms sender is created");
			// now find the stats and attache to wrapper so we can collect stats later
			ConnectionStats stats = new ConnectionStats();
			for (JmsSessionWrapper sw : existingSess) {
				JmsConfig cfg = sw.sessProp;
				if (StringUtil.equalIgnoreCase(cfg.getName(), queueName) 
						&& cfg.isProducing() && cfg.isQueue()) {
					stats = sw.getStats();
					break;
				}
			}
			return new QueueWrapper(session, sender, stats);
		} catch (JMSException e) {
			cat.error(e);
			throw e;
		}
	}
	
	/**
	 * queue wrapper
	 * @author biyu
	 *
	 */
	static class QueueWrapper {
		/** JMS queue session created */
		QueueSession sess;
		/** JMS publisher created */
		QueueSender sender;
		/** stats collector */
		ConnectionStats stats;
		QueueWrapper(QueueSession sess, QueueSender sender, ConnectionStats stats) {
			this.sess = sess;
			this.sender = sender;
			this.stats = stats;
		}
	}
	
	/**
	 * a working {@link Runnable} that does the publish work
	 * @author biyu
	 *
	 */
	static class QueueWorker implements Runnable {
		
		private static ThreadLocal<QueueWrapper> senders = new ThreadLocal<QueueWrapper>() {};

		private String queueName;
		private OutboundMessage msg;
		private ExceptionListener handler;
		
		QueueWorker(String queueName, OutboundMessage msg) {
			this(queueName, msg, null);
		}
		
		QueueWorker(String queueName, OutboundMessage msg, ExceptionListener handler) {
			this.queueName = queueName;
			this.msg = msg;
			this.handler = handler;
		}

		public void run() {
			QueueWrapper sender = senders.get();
			try {
				if (sender == null) {
					sender = MessageManager.createQueueSender(queueName);
					senders.set(sender);
				}
				TextMessage tmsg = sender.sess.createTextMessage(msg.getM_JMSContent());
				tmsg.setJMSCorrelationID(msg.getM_JMSCorrelationID());
				tmsg.setJMSDeliveryMode(msg.getM_JMSDeliveryMode());
				tmsg.setJMSPriority(msg.getM_JMSPriority());
				tmsg.setJMSExpiration(msg.getM_JMSExpiration());
				tmsg.setJMSTimestamp(msg.getM_JMSTimestamp());
				Map<String, Object> customHeaders = msg.getReadOnlyJMSProperties();
				if(customHeaders != null){
					for (Iterator<String> iter = customHeaders.keySet().iterator(); iter.hasNext(); ) {
						String key = (String)iter.next();
						Object value = customHeaders.get(key);
						tmsg.setObjectProperty(key, value);
					}
				}
				sender.sender.send(tmsg);
				sender.stats.onOutgoingMessage(msg);
				customHeaders = null;
			} catch (final JMSException e) {
				cat.error(e);
				if (handler != null) {
					// we don't want to block the publisher
					PriorityRoundRobinThreadPoolExecutor.getInstance().execute(
							new JmsQueueTask (new Runnable() {
								public void run() {
									handler.onException(new AsyncJmsException(e, msg));
								}
							})
					);
				}
			}
		}
	}

	/**
	 * A wrapper {@link MessageListener} to do some logging
	 * @author biyu
	 *
	 */
	static class MessageListenerWrapper implements MessageListener {
		private MessageListener listener;
		private JmsSessionWrapper sessWrapper;
		private ExecutorService executor;
		MessageListenerWrapper(JmsSessionWrapper sessWrapper, MessageListener listener) {
			this.listener = listener;
			this.sessWrapper = sessWrapper;
			executor = Executors.newSingleThreadExecutor();
		}
		MessageListenerWrapper(JmsSessionWrapper sessWrapper, MessageListener listener, ExecutorService executor) {
			this.listener = listener;
			this.sessWrapper = sessWrapper;
			this.executor = executor;
		}
		public void onMessage(final Message arg0) {
			
			executor.submit(new Runnable() {
				public void run() {
					// logging, stats gathering
					sessWrapper.getStats().onIncomingMessage(arg0);
					long s = System.nanoTime();
					listener.onMessage(arg0);
					// execution time
					sessWrapper.getStats().adjustExecutionDuration(System.nanoTime() - s);
				}
			});
		}
	}
	
	/**
	 * A wrapper {@link ExceptionListener} to do some logging and accomodate 
	 * all listeners created by each subscription
	 * @author biyu
	 *
	 */
	static class ExceptionListenerWrapper implements ExceptionListener {
		HashSet<ExceptionListener> listeners = new HashSet<ExceptionListener>();
		JmsSessionWrapper sessWrapper;
		ExceptionListenerWrapper(JmsSessionWrapper sessWrapper) {
			this.sessWrapper = sessWrapper;
		}
		public void onException(JMSException arg0) {
			try {
				for (ExceptionListener listener : listeners) {
					listener.onException(arg0);
				}
			} finally {
				sessWrapper.getStats().onException(arg0);
				cat.error("JMSException on jms conn with ErrorCdoe " + arg0.getErrorCode(), arg0);
			}
		}
		void addListener(ExceptionListener listener) {
			listeners.add(listener);
		}		
	}
	
	public long getDefExpiration() {
		return defExpiration;
	}
	Map<String, QueueConnection> getQueue2Conns() {
		return Collections.unmodifiableMap(queue2Conns);
	}
	Map<String, QueueConnection> getQueueConns() {
		return Collections.unmodifiableMap(queueConns);
	}
	Map<String, TopicConnection> getTopic2Conns() {
		return Collections.unmodifiableMap(topic2Conns);
	}
	Map<String, TopicConnection> getTopicConns() {
		return Collections.unmodifiableMap(topicConns);
	}

	Set<JmsSessionWrapper> getExistingSess() {
		return existingSess;
	}

}

