package org.lightj.locking;

import java.sql.SQLException;
import java.util.Date;

import org.lightj.dal.BaseDatabaseType;
import org.lightj.dal.ITransactional;
import org.lightj.dal.TransactionManager;
import org.lightj.locking.dal.GlobalSemaphoreDAO;
import org.lightj.locking.dal.GlobalSemaphoreDO;
import org.lightj.util.Log4jProxy;

/**
 * global semaphore manager
 * @author biyu
 *
 */
public final class GlobalSemaphoreManager {
	
	static final Log4jProxy log = Log4jProxy.getInstance(GlobalSemaphoreManager.class.getName());
	
	public static IGlobalSemaphore getGlobalSemaphore() {
		return new LockingModule().getGlobalSemaphore();
	}

	/**
	 * default global semaphore impl
	 * @author biyu
	 *
	 */
	public static class GlobalSemaphoreImpl implements IGlobalSemaphore {
		
		private BaseDatabaseType db;

		public GlobalSemaphoreImpl() {}
		public GlobalSemaphoreImpl(BaseDatabaseType db) {
			this.db = db;
		}
		
		public BaseDatabaseType getDb() {
			return db;
		}
		public void setDb(BaseDatabaseType db) {
			this.db = db;
		}

		/**
		 * This method will synchronize your code globally on an object.
		 * 
		 * The way synchronizeObject works is that it will lock a object across VMs in the database
		 * and will hold the lock until the transaction passed to it is not finished. 
		 * 
		 * @param semaphore
		 * @param context
		 * @throws Exception
		 */
		public final void synchronizeObject(final ISemaphoreObject semaphore, final ITransactional context) 
				throws RuntimeException, CriticalSectionException 
		{		
			//create the row for SemaphoreDO
			final GlobalSemaphoreDO semaphoreDO = new GlobalSemaphoreDO();
			semaphoreDO.setCreationDate(new Date());
			semaphoreDO.setLastModifiedDate(new Date());
			semaphoreDO.setSemaphoreKey(semaphore.getSemaphoreKey());

			/*This code will only work on Site controller database*/
			try{
				TransactionManager.execute(db, new ITransactional() {
					
					public void execute() throws Throwable {
						/* TRANSACTION */
						log.debug("Trying to acquire semaphore on " + " Thread id = " + Thread.currentThread());
						GlobalSemaphoreDAO.getInstance().lockSemaphore(db, semaphoreDO);
						log.debug("Acquired semaphore on " + " Thread id = " + Thread.currentThread());

						/*Inner transaction from another database enum.*/
						//call another TransactionalManager. - always executed against the site database.
						log.debug("Executing Critical Section " + " Thread id = " + Thread.currentThread());
						context.execute();
						log.debug("Execution of Critical Section completed, removing lock on semaphore for Thread id = " + Thread.currentThread());
						/*Transaction will automatically be committed.*/
					}
					
				});
				
			}catch(RuntimeException rte){
				throw rte;
			}catch (SQLException e){
				throw new RuntimeException(e);
			}catch (Exception etz){
				throw new CriticalSectionException(etz);
			}catch (Throwable t) {
				throw new CriticalSectionException(t.getMessage());
			}
		}
		
	}
	
}
