/*
 * Created on Jan 6, 2006
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package org.lightj.schedule;

import java.io.InvalidObjectException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicLong;

import org.lightj.dal.Locator;
import org.lightj.util.Log4jProxy;


/**
 * @author biyu
 *
 * This is just a simple base class provided for convenience
 * with common Schedulable's functionality. It can also be
 * used as noop schedulable. User is by no means forced 
 * to extend this class for their own Schedulables. Anyone with problem
 * extending it can just implement Schedulable interface with their own
 * concrete class. However it is very important to use 
 * {@link BaseSchedulable#generateKey()} in your own getKey() method,
 * since SchdulerFactory is expecting such implementation to be consistent.   
 */
@SuppressWarnings("rawtypes")
public class BaseSchedulable implements Schedulable {
	
	static Log4jProxy cat = Log4jProxy.getInstance(BaseSchedulable.class);
	
	/** used to generate a unique key for each schedulable */
	static AtomicLong keyGenerator = new AtomicLong(); 

	/** schedule */
	protected List<ISchedule> schedules;
	
	/** key */
	protected String key;
	
	/**
	 * Constructor
	 */
	public BaseSchedulable() {
		schedules = new ArrayList<ISchedule>();
	}

	/**
	 * Execute a job.
	 * @param ctx	persistent data for this job
	 */
	public void execute(IExecutionContext ctx) {
		// do nothing
	}
	
	/**
	 * Interrup an executing job. {@see SchedulerFactory#interrup(Schedulable)}
	 */
	public void interrupt() {
		// do nothing
	}
	
	/**
	 * Get list of {@link ISchedule}
	 * @return
	 */
	public List<ISchedule> getSchedules() {
		return schedules;
	}
	
	/**
	 * Add a new {@link ISchedule}
	 * @param schedule
	 */
	public void addSchedule(ISchedule schedule) {
		schedules.add(schedule);
	}
	
	/**
	 * Remove a {@link ISchedule}
	 * @param schedule
	 */
	public void removeSchedule(ISchedule schedule) {
		schedules.remove(schedule);
	}
	
	/**
	 * clear all schedules
	 *
	 */
	public void clearSchedule() {
		schedules.clear();
	}

	/////////////// Locatable interface //////////////
	/** get a key, or generate one if not exist */
	public synchronized String getKey() {
		if (key == null) {
			key = generateKey();
			BaseSchedulableLocator.getInstance().put(key, this);
		}
		return key;
	}
	
	/** take a key generated outside and set it */
	public synchronized void setKey(String key) {
		this.key = key;
	}

	/** get locator */
	public Class<? extends Locator> getLocator() {
		return BaseSchedulableLocator.class;
	}

	/** generate a unique key */
	public static synchronized final String generateKey() {
		return Long.toString(keyGenerator.incrementAndGet());
	}
	
	/**
	 * locator, simply put all locatables in a {@link ConcurrentMap} to be retrieved later
	 * 
	 * @author biyu
	 *
	 */
	public static class BaseSchedulableLocator implements Locator<BaseSchedulable> {
		
		/** singleton */
		static BaseSchedulableLocator me = new BaseSchedulableLocator();
		/** a map keep key to schedulable mapping */
		static ConcurrentMap<String, BaseSchedulable> map = new ConcurrentHashMap<String, BaseSchedulable>();
		/** get singleton */
		public static BaseSchedulableLocator getInstance() {
			return me;
		}
		/** constructor */
		private BaseSchedulableLocator() {}
		/** find schedulable from a key */
		public BaseSchedulable findByKey(String key) throws InvalidObjectException {
			return map.get(key);
		}
		/** add key to schedulable mapping */
		void put(String key, BaseSchedulable schedulable) {
			map.put(key, schedulable);
		}
		/** remove a mapping */
		public void remove(String key) {
			map.remove(key);
		}
	}
	
}
