/*
 * Created on Jan 4, 2006
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package org.lightj.schedule;

import java.io.Serializable;
import java.util.Date;
import java.util.Enumeration;

import org.quartz.CronTrigger;
import org.quartz.SimpleTrigger;

/**
 * @author biyu
 * Schedulable interface
 */
@SuppressWarnings("rawtypes")
public interface ISchedule extends Serializable {
	
	/**
	 * Repeat times constant, {@see #executeRecur(Date, Date, int, long)}
	 */
	public static final int REPEAT_INDEFINITELY = -1;

	/**
	 * Used to setup a continuous schedule , {@see #executeRecur(Date, Date, int, long)}
	 */
	public static final Date NO_END_DATE = null;
	
	/**
	 * Default misfire instruction {@see #setMisFireInstruction(int)}
	 */
	public static final int MISFIRE_INSTRUCTION_SMART_POLICY = 0;
	
	/**
	 * misfire instruction constant for simple job see {@link #setSimpleMisFireInstruction(int)}
	 */
	public static final int SIMPLE_MISFIRE_INSTRUCTION_FIRE_NOW = 
		SimpleTrigger.MISFIRE_INSTRUCTION_FIRE_NOW;
	public static final int SIMPLE_MISFIRE_INSTRUCTION_RESCHEDULE_NOW_WITH_EXISTING_REPEAT_COUNT = 
		SimpleTrigger.MISFIRE_INSTRUCTION_RESCHEDULE_NOW_WITH_EXISTING_REPEAT_COUNT;
	public static final int SIMPLE_MISFIRE_INSTRUCTION_RESCHEDULE_NOW_WITH_REMAINING_REPEAT_COUNT = 
		SimpleTrigger.MISFIRE_INSTRUCTION_RESCHEDULE_NOW_WITH_REMAINING_REPEAT_COUNT;
	public static final int SIMPLE_MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_REMAINING_COUNT = 
		SimpleTrigger.MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_REMAINING_COUNT;
	public static final int SIMPLE_MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_EXISTING_COUNT = 
		SimpleTrigger.MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_EXISTING_COUNT;
	public static final int MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_EXISTING_COUNT = 
		SimpleTrigger.MISFIRE_INSTRUCTION_RESCHEDULE_NEXT_WITH_EXISTING_COUNT;
	
	/**
	 * misfire instruction constant for cron job see {@link #setCronMisFireInstruction(int)}
	 */
	public static final int CRON_MISFIRE_INSTRUCTION_FIRE_ONCE_NOW = 
		CronTrigger.MISFIRE_INSTRUCTION_FIRE_ONCE_NOW;
	public static final int CRON_MISFIRE_INSTRUCTION_DO_NOTHING = 
		CronTrigger.MISFIRE_INSTRUCTION_DO_NOTHING;

	
	/**
	 * Execute once at executionTime, if null, execute immediately
	 * @param executionTime
	 * @throws ScheduleException
	 */	
	public void executeAt(Date executionTime) throws ScheduleException;
	
	/**
	 * Execute repeatly from startTime to endTime with interval, 
	 * at most repeat repeatCount times. This method round all times to second.
	 * @param startTime		start execution, immediately if null
	 * @param endTime		end execution
	 * @param repeatCount	max repeat times
	 * @param intervalSec	repeat interval in second
	 * @throws ScheduleException
	 */
	public void executeRecur(Date startTime, Date endTime, int repeatCount, long intervalSec) throws ScheduleException;
	
	/**
	 * Execute based on a cron expression. This method round all times to second.
	 * @param startTime		start time when cron expression is valid
	 * @param endTime		end time when cron expression is valid
	 * @param cronExpression
	 * @throws ScheduleException
	 */ 
	public void executeCron(Date startTime, Date endTime, String cronExpression) throws ScheduleException;
	
	/**
	 * Execute based on a cron expression with a randomized increment in seconds.
	 * The start and end times are rounded to the second.
	 * @param startTime
	 * @param endTime
	 * @param cronExpression
	 * @param upperBoundCronSecondIncrement used as the upper bound to generate a random number
	 * 									for the cron's second increment
	 * @throws ScheduleException
	 */
	public void executeCron(Date startTime, Date endTime, String cronExpression, int upperBoundCronSecondIncrement)
		throws ScheduleException;
	
	/**
	 * Don't execute during specified period. This mothed round all times to second.
	 * @param excludeStart	first occurance of the excluded period start
	 * @param excludeEnd	first occurance of the excluded period end
	 * @param intervalSec	should excluded period repeat, repeat interval, 0 for no repeat
	 * @param endDate		end date when this excluion rule stop being followed
	 * @throws ScheduleException
	 */
	public void executeExclude(Date excludeStart, Date excludeEnd, long intervalSec, Date endDate) throws ScheduleException;
	
	/**
	 * Pause this schedule.
	 * @throws ScheduleException
	 */
	public void pause() throws ScheduleException;
	
	/**
	 * Resume this schedule.
	 * @throws ScheduleException
	 */	
	public void resume() throws ScheduleException;
	
	/**
	 * Clear all associated {@link IExecutionContext}, include and exclude execution time.
	 * Effetively the same as construct a brand new schedule.
	 */
	public void clear();
	
	/**
	 * @return next execution time, null if no more execution.
	 */
	public Date getNextExecutionTime();
	
	/**
	 * @return final execution time, null if repeat infinitely.
	 */
	public Date getFinalExecutionTime();
	
	/**
	 * @return previous execution time, null if hasn't been executed yet.
	 */
	public Date getPrevExecutionTime();
	
	/**
	 * @return {@link java.util.Enumeration} of future execution times (could be endless).
	 */
	public Enumeration getExecutionTimes();
	
	/**
	 * @return {@link IExecutionContext} associated with this schedule
	 */
	public IExecutionContext getContext();
	
	/**
	 * Set misfire instruction 
	 * @param misfireInstruction
	 */
	public void setSimpleMisFireInstruction(int misfireInstruction);
	
	/**
	 * Get misfire instruction
	 * @return
	 */
	public int getSimpleMisFireInstruction();
	
	/**
	 * Set misfire instruction 
	 * @param misfireInstruction
	 */
	public void setCronMisFireInstruction(int misfireInstruction);
	
	/**
	 * Get misfire instruction
	 * @return
	 */
	public int getCronMisFireInstruction();

	/**
	 * User friendly description for this schedule.
	 * @param description
	 */
	public void setDescription(String description);

	/**
	 *@return User friendly description for this schedule.
	 */
	public String getDescription();
	
	/**
	 * Whether this schedule has been scheduled by scheduler.
	 * {@see SchedulerFactory#schedule(Schedulable)} and {@see SchedulerFactory#reschedule(Schedulable)}
	 * @return
	 */
	public boolean isScheduled();
	
}
