package org.lightj.schedule;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.sql.SQLException;
import java.util.Properties;

import org.lightj.RuntimeContext;
import org.lightj.dal.BaseDatabaseType;
import org.lightj.dal.ConnectionHelper;
import org.lightj.dal.DatabaseModule;
import org.lightj.dal.HsqlDatabaseType;
import org.lightj.initialization.BaseInitializable;
import org.lightj.initialization.BaseModule;
import org.lightj.initialization.InitializationException;
import org.lightj.util.FileUtils;
import org.lightj.util.Log4jProxy;
import org.lightj.util.StringUtil;
import org.quartz.Scheduler;
import org.quartz.impl.StdSchedulerFactory;

/**
 * place to intialize your scheduler, take array of {@link SchedulerTypeEnum}
 * 
 * {@link SchedulerTypeEnum#SCHEDULER_PERSISTENT} persistent scheduler 
 * need a config file quartz.persistent.properties in its config root {@link RuntimeContext#getConfigRoot()}
 * persistent storage set by {@link SchedulerConfigBean#setPersistentSchedulerDb(BaseDatabaseType)} prior to initialize this module 
 * 
 * {@link SchedulerTypeEnum#SCHEDULER_TRANSIENT} transient scheduler
 * scheduler threadpool size can be set by {@link SchedulerConfigBean#setTransientSchedulerThreadSize(int)} prior to initialize this module
 * 
 * @author biyu
 *
 */
public class PersistentSchedulerModule {
	
	/** logger */
	static Log4jProxy logger = Log4jProxy.getInstance(PersistentSchedulerModule.class);

	/** singleton */
	private static PersistentSchedulerModuleInner s_Module = null;
	
	public PersistentSchedulerModule() {
		init();
	}
	
	private synchronized void init() {
		if (s_Module == null) s_Module = new PersistentSchedulerModuleInner();
	}
	
	/**
	 * data has to be set before the module initialized
	 * @param dbEnum
	 */
	public PersistentSchedulerModule setDatabase(BaseDatabaseType dbEnum) {
		s_Module.addDependentModule(new DatabaseModule().addDatabases(dbEnum).getModule());
		s_Module.dbEnum = dbEnum;
		return this;
	}
	
	public BaseModule getModule() {
		return s_Module;
	}

	
	private class PersistentSchedulerModuleInner extends BaseModule {
		
		private BaseDatabaseType dbEnum;
		
		/** constructor, initialize */
		private PersistentSchedulerModuleInner() {
			super();

			addInitializable(new BaseInitializable() {
				
				@Override
				protected void initialize() {
					
					if (dbEnum == null || !dbEnum.isShared()) { 
						throw new InitializationException("persistent scheduler requires a shared database");
					}

					URL configRoot = RuntimeContext.getConfigRoot();
					try {
						
						if (configRoot != null) {
							if (dbEnum instanceof HsqlDatabaseType) {
								setupTables(new URL(configRoot, "hsqldb_quartz.sql"));
							}
							
							QuartzConnectionProvider.setDbEnum(dbEnum);
							initPersistentScheduler(new URL(configRoot, "quartz.persistent.properties"));
						} 
						else {
							throw new InitializationException("Please specify config root where quartz.persistent.properties can be found first");
						}
					} catch (ScheduleException e) {
						logger.fatal("Scheduler failed to initialize", e);
						throw new InitializationException(e);
					} catch (MalformedURLException e) {
						logger.fatal("Wrong path to persistent scheduler config file", e);
						throw new InitializationException(e);
					} catch (IOException e) {
						logger.fatal("Error setting up the tables", e);
						throw new InitializationException(e);
					} catch (SQLException e) {
						logger.fatal("Error setting up the tables", e);
						throw new InitializationException(e);
					}
				}

				@Override
				protected void shutdown() {
					try {
						SchedulerFactory.getInstance(SchedulerTypeEnum.SCHEDULER_PERSISTENT).shutdown();
					} catch (ScheduleException e) {
						logger.error("Failed to shutdown scheduler " + SchedulerTypeEnum.SCHEDULER_PERSISTENT, e);
					}
				}

			});
		}
		
		/**
		 * init persistent scheduler
		 * @param config
		 * @param startNow
		 * @throws ScheduleException
		 */
		private final void initPersistentScheduler(URL configFile) throws ScheduleException
		{
			StdSchedulerFactory factory;
			try {
				Properties props = new Properties();
				props.load(configFile.openStream());
				factory = new StdSchedulerFactory(props);

				Scheduler scheduler = factory.getScheduler();
				SchedulerFactory.addInstance(SchedulerFactory.SCHEDULER_PERSISTENT, scheduler, false);
				
				scheduler.start();
				logger.info("Persistent scheduler started...");

			} catch (Exception e) {
				throw new ScheduleException(e);
			}
		}
		
		private void setupTables(URL sqlFile) throws IOException, SQLException {
			String queryStr = FileUtils.readStream(sqlFile.openStream());
			String[] queries = queryStr.split(";");
			for (String query : queries) {
				if (!StringUtil.isNullOrEmptyAfterTrim(query)) {
					ConnectionHelper.executeUpdate(dbEnum, query);
				}
			}
		}
	}
	
}
