/*
 * Created on Jan 4, 2006
 *
 */
package org.lightj.schedule;

import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.UUID;

import org.apache.log4j.Level;
import org.lightj.dal.LocatorUtil;
import org.lightj.util.DateUtil;
import org.lightj.util.Log4jProxy;
import org.lightj.util.NetUtil;
import org.lightj.util.StringUtil;
import org.lightj.util.nanoxml.XMLElement;
import org.quartz.CronTrigger;
import org.quartz.InterruptableJob;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.SchedulerException;
import org.quartz.SimpleTrigger;
import org.quartz.StatefulJob;
import org.quartz.Trigger;
import org.quartz.TriggerUtils;

/**
 * @author biyu
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
@SuppressWarnings({"rawtypes","unchecked"})
public class QuartzScheduler extends SchedulerFactory { 
	
	public static final String CLEANUPAFTERFIRE		= "CLEANUPAFTERFIRE";
	
	private static final Log4jProxy _log = Log4jProxy.getLogger(QuartzScheduler.class);

	private static final Random s_random = new Random();
	
	/**
	 * whether the scheduler needs db connection to work
	 */
	boolean isTransient;
	
	/**
	 * Quartz scheduler
	 */
	org.quartz.Scheduler scheduler;
	
	/**
	 * Format date
	 */
	static final SimpleDateFormat fmt = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	
	/**
	 * schedule a schedulable to be runned by Quartz scheduler
	 */
	public void schedule(Schedulable schedulable) throws ScheduleException {
		List<ISchedule> schedules = schedulable.getSchedules();
		if (schedules != null) {
			for (int i = 0, size = schedules.size(); i < size; i++) {
				try {
					QuartzSchedule schedule = (QuartzSchedule) schedules.get(i);
					if (schedule.isScheduled()) {
						throw new UnsupportedOperationException("job already scheduled, should use reschedule instead");
					}
					String jobName = Long.toString(System.nanoTime()) + '.' + schedulable.getKey();
					if (!this.isTransient && jobName.length() > 80) {
						logger.log(Level.WARN, "schedule job name too long, max length 80 character " + jobName);
					}
					jobName = StringUtil.trimToLength(jobName, 80);
					String jobGroup = schedulable.getClass().getName();
					JobDetail job = new JobDetail(jobName, jobGroup, QuartzJob.class);
					job.setDescription(schedule.getDescription());
					
					QuartzExecutionContext ctx = (QuartzExecutionContext) schedule.getContext();
					List excludedTimes = schedule.getExcludedTime(); 
					if (excludedTimes != null && !excludedTimes.isEmpty()) {
						ctx.setStringParm(QuartzSchedule.EXCLUDED_TIME, StringUtil.join(excludedTimes, ","));
					}
					// there is only one trigger, cleanup candidate, if JobCleanupListener is enabled, 
					// this job is going to be cleaned up after completion
					if (schedule.getTriggers().size() == 1) {
						ctx.setBooleanParm(CLEANUPAFTERFIRE, true);
					}
					else {
						ctx.setBooleanParm(CLEANUPAFTERFIRE, false);
					}
					ctx.setStringParm(IExecutionContext.SCHEDULABLE_ID, LocatorUtil.getLocatableString(schedulable));
					// save data
					job.setJobDataMap(ctx.getData());
					scheduler.addJob(job, true);
					// keep a reference in case we need to delete it later
					schedule.setJob(job);
					// create triggers
					int tcounter = 0;
					for (Iterator iter = schedule.getTriggers().iterator(); iter.hasNext(); ) {
						Trigger t = (Trigger) iter.next();
						t.setGroup(jobName);
						t.setJobName(jobName);
						t.setJobGroup(jobGroup);
						t.setName(jobName + '.' + Integer.toString(tcounter++));
						// we don't have flexiblity to handle misfire instruction yet, always use the default(smart) instruction
						if (t instanceof SimpleTrigger) {
							t.setMisfireInstruction(schedule.getSimpleMisFireInstruction());
						}
						else if (t instanceof CronTrigger) {
							t.setMisfireInstruction(schedule.getCronMisFireInstruction());
						}
						scheduler.scheduleJob(t); 
					}
				}
				catch (SchedulerException se) {
					_log.error("Failed to schedule " + schedulable.getClass().getName() + "," +schedulable.getKey(), se);
					throw new ScheduleException(se);
				}
			}			
		}
	}
	
	/**
	 * cancel a schedulable
	 */
	public void cancel(Schedulable schedulable) throws ScheduleException {
		// load schedule if that's empty
		if (schedulable.getSchedules().isEmpty()) {
			List<ISchedule> schedules = getSchedules(schedulable);
			schedulable.getSchedules().addAll(schedules);
		}
		for (Iterator iter = schedulable.getSchedules().iterator(); iter.hasNext();) {
			try {
				QuartzSchedule s = (QuartzSchedule) iter.next();
				if (!s.isScheduled()) continue;
				JobDetail job = s.getJob();
				if (job != null) {
					scheduler.deleteJob(job.getName(), job.getGroup());
					s.setJob(null);
				}
			}
			catch (SchedulerException se) {
				_log.error("Failed to reschedule", se);
				throw new ScheduleException(se);
			}
		}
	}
	
	/**
	 * pause a non-executing schedulable
	 */
	public void pause(Schedulable schedulable) throws ScheduleException {
		// load schedule if that's empty
		if (schedulable.getSchedules().isEmpty()) {
			List<ISchedule> schedules = getSchedules(schedulable);
			schedulable.getSchedules().addAll(schedules);
		}
		for (Iterator iter = schedulable.getSchedules().iterator(); iter.hasNext();) {
			try {
				JobDetail job = ((QuartzSchedule) iter.next()).getJob();
				scheduler.pauseJob(job.getName(), job.getGroup());
			} catch (SchedulerException e) {
				_log.error("failed to pause(Schedulable)", e);
				throw new ScheduleException(e);
			}
		}
	}
	
	/**
	 * resume a schedulable
	 */
	public void resume(Schedulable schedulable) throws ScheduleException {
		// load schedule if that's empty
		if (schedulable.getSchedules().isEmpty()) {
			List<ISchedule> schedules = getSchedules(schedulable);
			schedulable.getSchedules().addAll(schedules);
		}
		for (Iterator iter = schedulable.getSchedules().iterator(); iter.hasNext();) {
			try {
				JobDetail job = ((QuartzSchedule) iter.next()).getJob();
				scheduler.resumeJob(job.getName(), job.getGroup());
			} catch (SchedulerException e) {
				_log.error("failed to resume(Schedulable)", e);
				throw new ScheduleException(e);
			}
		}
	}

	/**
	 * interrupt an executing schedulable
	 * @param s
	 * @throws ScheduleException
	 */	
	public void interrupt(Schedulable schedulable) throws ScheduleException {
		// load schedule if that's empty
		if (schedulable.getSchedules().isEmpty()) {
			List<ISchedule> schedules = getSchedules(schedulable);
			schedulable.getSchedules().addAll(schedules);
		}
		for (Iterator iter = schedulable.getSchedules().iterator(); iter.hasNext();) {
			try {
				JobDetail job = ((QuartzSchedule) iter.next()).getJob();
				scheduler.interrupt(job.getName(), job.getGroup());
			} catch (SchedulerException e) {
				_log.error("failed to resume(Schedulable)", e);
				throw new ScheduleException(e);
			}
		}
	}
	
	/**
	 * Clear all previous jobs and reschedule
	 * @param schedulable
	 * @throws ScheduleException
	 */
	public void reschedule(Schedulable schedulable) throws ScheduleException {
		cancel(schedulable);
		schedule(schedulable);
	}
	
	/**
	 */
	public void setThirdPartyScheduler(Object s) {
		scheduler = (org.quartz.Scheduler) s;
	}
	
	/**
	 * @param isTransient
	 */
	public void setTransient(boolean isTransient) {
		this.isTransient = isTransient;
	}

	/**
	 * create a Quartz specific ISchedule
	 */
	public ISchedule createSchedule(String description) throws ScheduleException 
	{
		 return new QuartzSchedule(description);
	}
	
	/**
	 * pause the whole scheduler
	 */
	public void pause() throws ScheduleException {
		try {
			scheduler.pauseAll();
		}
		catch (SchedulerException e) {
			_log.error("failed to pause()", e);
			throw new ScheduleException(e);
		}
	}
	
	/**
	 * resume the whole scheduler
	 */
	public void resume() throws ScheduleException {
		try {
			scheduler.resumeAll();
		}
		catch (SchedulerException e) {
			_log.error("failed to resume()", e);
			throw new ScheduleException(e);
		}
	}
	
	public void start() throws ScheduleException {
		try {
			scheduler.start();
		} catch (SchedulerException e) {
			_log.error("failed to start()", e);
			throw new ScheduleException(e);
		}
	}
	
	public void shutdown() throws ScheduleException {
		try {
			scheduler.shutdown();
		} catch (SchedulerException e) {
			_log.error("failed to shutdown()", e);
			throw new ScheduleException(e);
		}
	}
	
	/**
	 * pause a whole schedulable group
	 */
	public void pause(Class schedulableKlass) throws ScheduleException {
		try {
			if (!Schedulable.class.isAssignableFrom(schedulableKlass))
				throw new IllegalArgumentException("can't pause a non Schedulable class");
			scheduler.pauseJobGroup(schedulableKlass.getName());
		}
		catch (SchedulerException e) {
			_log.error("failed to pause(Class)", e);
			throw new ScheduleException(e);
		}
	}
	
	/**
	 * resume a whole scheduler group
	 */
	public void resume(Class schedulableKlass) throws ScheduleException {
		try {
			if (!Schedulable.class.isAssignableFrom(schedulableKlass))
				throw new IllegalArgumentException("can't resume a non Schedulable class");
			scheduler.resumeJobGroup(schedulableKlass.getName());
		}
		catch (SchedulerException e) {
			_log.error("failed to resume(Class)", e);
			throw new ScheduleException(e);
		}
	}
	
	/**
	 * Quartz wrapper for schedulable, it's the one who is scheduled in Quartz
	 * scheduler and do the real job.
	 * 
	 * @author biyu
	 */
	public static final class QuartzJob implements StatefulJob, InterruptableJob
	{
		/**
		 * Quartz scheduler need empty constructor
		 *
		 */
		public QuartzJob() {}

		/**
		 * see <code>{@link org.quartz.StatefulJob}</code>
		 */
		public void execute(JobExecutionContext context) throws JobExecutionException 
		{
			try {
				JobDataMap jobDataMap = context.getJobDetail().getJobDataMap();
				Date executionTime = context.getFireTime();
				// skip if fall into excluded time
				if (jobDataMap.containsKey(QuartzSchedule.EXCLUDED_TIME)) {
					if (isExcluded(executionTime, jobDataMap.getString(QuartzSchedule.EXCLUDED_TIME))) {
						return;
					}
				}
				// find out concrete schedulable and instantiate it.
				// changed on 6/6/06, keep the old code for compatibility
				Schedulable schedulable = null;
				String locatableStr = jobDataMap.getString(IExecutionContext.SCHEDULABLE_ID);
				schedulable = (Schedulable) LocatorUtil.findByLocatableString(locatableStr);
				QuartzExecutionContext ctx = new QuartzExecutionContext(jobDataMap);

				_log.debug(" name " + schedulable.getClass().getName());

				schedulable.execute(ctx);
				
				_log.debug(" scheduler trigger completed ");
			}
			catch (Exception e) {
				_log.error(e);
				throw new JobExecutionException(e);
			}
		}
		
		/**
		 * see <code>{@link org.quartz.InterruptableJob}</code>
		 */
		public void interrupt() {
			/* Fix for FindBugs Violation              
			 * Violation Name: UWF_UNWRITTEN_FIELD
			 * Category: CORRECTNESS
			 */
//			if (schedulable != null) schedulable.interrupt();
		}

		/**
		 * wether an execution time should be excluded
		 * @param d
		 * @param excludeStr
		 * @return
		 */
		private boolean isExcluded(Date d, String excludeStr) {
			if (excludeStr == null)
				return false;
			long dmilli = d.getTime();
			List excludes = StringUtil.split(excludeStr, ",");
			for (int i = 0, size = excludes.size(); i < size; i += 4) {
				long start = Long.parseLong((String) excludes.get(i));
				long end = Long.parseLong((String) excludes.get(i + 1));
				long interval = Long.parseLong((String) excludes.get(i + 2));
				long excludeEnd = Long.parseLong((String) excludes.get(i + 3));
				// exclusion period haven't start yet, or end already
				if (start > dmilli || (excludeEnd > 0 && dmilli > excludeEnd))
					continue;
				if (dmilli >= start && dmilli <= end) {
					return true;
				}
				if (interval > 0) {
					long distance = end - start + (interval * 1000L);
					long m = (dmilli - start) / distance;
					long nend = end + m * distance;
					if (dmilli <= nend)
						return true;
				}
			}
			return false;
		}
	}


	/**
	 * IExecutionContext implementation for Quartz, wrap Quartz JobDataMap
	 * 
	 * @author biyu
	 */
	public static class QuartzExecutionContext implements IExecutionContext {
	
		private static final long serialVersionUID = 3272333911295610737L;
		private JobDataMap data;
		
		QuartzExecutionContext() {
			data = new JobDataMap();
		}
	
		QuartzExecutionContext(JobDataMap data) {
			this.data = data;
		}
		
		public int getIntParm(String name, int defVal) {
			return (data.containsKey(name) ? data.getInt(name) : defVal);
		}
	
		public void setIntParm(String name, int val) {
			data.put(name, val);
		}
	
		public long getLongParm(String name, long defVal) {
			return (data.containsKey(name) ? data.getLong(name) : defVal); 
		}
	
		public void setLongParm(String name, long val) {
			data.put(name, val);
		}
	
		public double getDoubleParm(String name, double defVal) {
			return (data.containsKey(name) ? data.getDouble(name) : defVal);
		}
	
		public void setDoubleParm(String name, double val) {
			data.put(name, val);
		}
	
		public boolean getBooleanParm(String name, boolean defVal) {
			return (data.containsKey(name) ? data.getBoolean(name) : defVal);
		}
	
		public void setBooleanParm(String name, boolean val) {
			data.put(name, val);
		}
	
		public Date getDateParm(String name, boolean includeTime, Date defVal) {
			if (data.containsKey(name)) {
				try {
					Date d = fmt.parse(data.getString(name));
					if (!includeTime) {
						Calendar cal = Calendar.getInstance();
						cal.setTime(d);
						cal.set(Calendar.HOUR, 0);
						cal.set(Calendar.MINUTE, 0);
						cal.set(Calendar.SECOND, 0);
						cal.set(Calendar.MILLISECOND, 0);
						d = cal.getTime();
					}
					return d;
				}
				catch (Exception e) {
					return defVal;
				}
			}
			return defVal;
		}
	
		public void setDateParm(String name, Date val) {
			data.put(name, fmt.format(val));
		}
	
		public String getStringParm(String name, String defVal) {
			return (data.containsKey(name) ? data.getString(name) : defVal);
		}
	
		public void setStringParm(String name, String val) {
			data.put(name, val);
		}
	
		public Serializable getParm(String name) {
			return (data.containsKey(name) ? (Serializable) data.get(name) : null);
		}
		
		public void setParm(String name, Serializable val) {
			data.put(name, val);
		}
	
		public void removeParm(String name) {
			data.remove(name);
		}
		
		public void clear() {
			data.clear();
		}
	
		public Map getAll() {
			return Collections.unmodifiableMap(data); 
		}
		
		public void putAll(Map all) {
			for (Iterator iter = all.keySet().iterator(); iter.hasNext(); ) {
				Object key = iter.next();
				data.put(key, all.get(key));
			}
		}
		
		JobDataMap getData() {
			return data;
		}
	}

	/**
	 * Quartz implementation of ISchedule
	 * @author biyu
	 */
	public final class QuartzSchedule implements ISchedule {
		
		private static final long serialVersionUID = 1826784546023655558L;

		/**
		 * used internally by QuartzJob to exclude certain date range from execution
		 */
		public static final String EXCLUDED_TIME = "_EXCLUDED_TIME";
		
		/**
		 * temporary name used for unscheduled triggers, 
		 * because they required a name in Quartz, replaced by real name when scheduled
		 */
		private static final String UNSCHEDULED_NAME = "UNSCHEDULED";

		/**
		 * reference to a scheduled Quartz job
		 */
		private JobDetail job;
		
		/**
		 * user friendly description for the job and this schedule
		 */
		private String description;
		
		/**
		 * execution context
		 */
		private QuartzExecutionContext ctx;
		
		/**
		 * Misfire instruction, see <code>{@link ISchedule }</code> 
		 */
		private int simpleMisfireInstruction = MISFIRE_INSTRUCTION_SMART_POLICY;
		
		/**
		 * misfire isntruction, see <code>{@link ISchedule}</code>
		 */
		private int cronMisfireInstruction = MISFIRE_INSTRUCTION_SMART_POLICY;

		/**
		 * List of <code>{@link org.quartz.Trigger}</code>
		 */
		private List<Trigger> triggers;
		
		/**
		 * list of excluded date range that this job should not be executed
		 */
		private List<String> excludedTimes;

		/////////////// constructors ////////////////
		/**
		 * create a new QuartzSchedule
		 */
		QuartzSchedule(String description) {
			this.description = description;
			ctx = new QuartzExecutionContext();
			triggers = new ArrayList<Trigger>();
			excludedTimes = new ArrayList<String>();
		}
		
		/**
		 * reconstruct a QuartzSchedule giving a Quartz job
		 * @param job
		 * @throws SchedulerException
		 */
		QuartzSchedule(JobDetail job) throws SchedulerException {
			this.job = job;
			this.description = job.getDescription();
			this.ctx = new QuartzExecutionContext(job.getJobDataMap());
			this.triggers = Arrays.asList(scheduler.getTriggersOfJob(job.getName(), job.getGroup()));
			String exp = ctx.getStringParm(EXCLUDED_TIME, null);
			if (exp != null) {
				excludedTimes = StringUtil.split(exp, ", ");
			}
		}
		
		Collection getTriggers() {
			return triggers;
		}
		
		List<String> getExcludedTime() {
			return excludedTimes;
		}
		
		JobDetail getJob() {
			return job;
		}
		
		void setJob(JobDetail job) {
			this.job = job;
		}
		
		/////////////////// ISchedule interface ///////////////
		
		public boolean isScheduled() {
			return (job != null);
		}
		
		public IExecutionContext getContext() {
			return ctx;
		}
		
		public String getDescription() {
			return description;
		}

		public void setDescription(String string) {
			description = string;
		}
		
		
		public void executeAt(Date executionTime) throws ScheduleException {
			executionTime = TriggerUtils.getEvenSecondDate(executionTime);
			SimpleTrigger t = new SimpleTrigger(getKey(), UNSCHEDULED_NAME, UNSCHEDULED_NAME, UNSCHEDULED_NAME, executionTime, null, 0, 0);
			t.setDescription("Execute once at " + fmt.format(executionTime)+ " by "+NetUtil.getMyHostName());
			t.setMisfireInstruction(simpleMisfireInstruction);
			triggers.add(t);
		}
	
		public void executeRecur(Date startTime, Date endTime, int repeatCount, long intervalSec)
			throws ScheduleException 
		{
			startTime = TriggerUtils.getEvenSecondDate(startTime);
			if (endTime != null) endTime = TriggerUtils.getEvenSecondDate(endTime);
			SimpleTrigger t = new SimpleTrigger(getKey(), UNSCHEDULED_NAME, UNSCHEDULED_NAME, UNSCHEDULED_NAME, startTime, endTime, repeatCount, intervalSec * 1000L);
			t.setDescription("Execute " + ((repeatCount > REPEAT_INDEFINITELY) ? Integer.toString(repeatCount) : "infinite") 
						+ " times from " + ((startTime != null) ? fmt.format(startTime) : "now")
						+ " to " + ((endTime != null) ? fmt.format(endTime) : "infinite")
						+ " for every " + intervalSec + " second");
			t.setMisfireInstruction(simpleMisfireInstruction);
			triggers.add(t);
		}
		
		private static final String CRON_FIELD_SEPERATOR = " ";
		public void executeCron(Date startTime, Date endTime, String cronExpression, int upperBoundCronSecondIncrement)
			throws ScheduleException
		{
			String expression = cronExpression;
			if (upperBoundCronSecondIncrement > 0) {
				// will override the provided second field from expression with a randomly
				// generated second value based on the given parameter
				int random = s_random.nextInt(upperBoundCronSecondIncrement);
				StringBuffer sb = new StringBuffer(20);
				int postSecondField = cronExpression.indexOf(CRON_FIELD_SEPERATOR);
				if (postSecondField == -1) {
					throw new ScheduleException(
						"Unable to execute cron because of an invalid cron expression: "+cronExpression);
				}
				sb.append(random).append(cronExpression.substring(postSecondField));
				expression = sb.toString();
				_log.error("Executing cron expression: "+expression);
			}
			executeCron(startTime, endTime, expression);
		}
	
		public void executeCron(Date startTime, Date endTime, String cronExpression)
			throws ScheduleException 
		{
			try {
				startTime = TriggerUtils.getEvenSecondDate(startTime);
				if (endTime != null) endTime = TriggerUtils.getEvenSecondDate(endTime);
				CronTrigger t = new CronTrigger(getKey(), UNSCHEDULED_NAME, UNSCHEDULED_NAME, UNSCHEDULED_NAME, startTime, endTime, cronExpression);
				t.setDescription("Execute from " + ((startTime != null) ? fmt.format(startTime) : "now")
								+ " to " + ((endTime != null) ? fmt.format(endTime) : "infinite")
								+ " based on cron expression " + cronExpression);
				t.setMisfireInstruction(CronTrigger.MISFIRE_INSTRUCTION_DO_NOTHING);
				triggers.add(t);
			}
			catch (Exception e) {
				throw new ScheduleException(e);
			}
		}
		
		public void executeExclude(Date excludeStart, Date excludeEnd, long intervalSec, Date endDate) 
			throws ScheduleException
		{
			if (excludeStart == null || excludeEnd == null) {
				throw new IllegalArgumentException("exclude start and end time cannot be null");
			}
			
			excludedTimes.add(Long.toString(TriggerUtils.getEvenSecondDateBefore(excludeStart).getTime()));
			excludedTimes.add(Long.toString(TriggerUtils.getEvenSecondDate(excludeEnd).getTime()));
			excludedTimes.add(Long.toString(intervalSec));
			excludedTimes.add((endDate != null) ? Long.toString(TriggerUtils.getEvenSecondDate(endDate).getTime()) : "0");
		}
	
		public void pause() throws ScheduleException {
			try  {
				scheduler.pauseJob(job.getName(), job.getGroup());
			}
			catch (Exception e) {
				_log.error("Failed to pause", e);
				throw new ScheduleException(e);
			}
		}
	
		public void resume() throws ScheduleException {
			try {
				scheduler.resumeJob(job.getName(), job.getGroup());
			}
			catch (Exception e) {
				_log.error("Failed to resume", e);
				throw new ScheduleException(e);
			}
		}
		
		public void clear() {
			ctx = new QuartzExecutionContext();
			triggers.clear();
			excludedTimes.clear();
		}
	
		/**
		 * @return null means no execution in future
		 */
		public Date getNextExecutionTime() {
			Date nextExecutionTime = null;
			for (Iterator iter = triggers.iterator(); iter.hasNext(); ) {
				Trigger t = (Trigger) iter.next();
				nextExecutionTime = DateUtil.min(nextExecutionTime, t.getNextFireTime(), false);
			}
			return nextExecutionTime;
		}
	
		/**
		 * @return null means repeat infinitely
		 */
		public Date getFinalExecutionTime() {
			// use some really old date to start
			Date finalExecutionTime = new Date(System.currentTimeMillis() - (365 * 24 * 60 * 60 * 1000L));
			for (Iterator iter = triggers.iterator(); iter.hasNext(); ) {
				Trigger t = (Trigger) iter.next();
				finalExecutionTime = DateUtil.max(finalExecutionTime, t.getFinalFireTime(), true);
			}
			return finalExecutionTime;
		}
	
		/**
		 * @return null means no execution before
		 */
		public Date getPrevExecutionTime() {
			Date prevExecutionTime = null;
			for (Iterator iter = triggers.iterator(); iter.hasNext(); ) {
				Trigger t = (Trigger) iter.next();
				prevExecutionTime = DateUtil.max(prevExecutionTime, t.getPreviousFireTime(), false);
			}
			return prevExecutionTime;
		}
		
		/**
		 * 
		 * @return Enumeration of {@link java.util.Date} of future execution time.
		 */
		public Enumeration getExecutionTimes() {
			return new ExecutionTimeEnumeration();
		}
		
		public String toString() {
			StringBuffer buf = new StringBuffer();
			for (Iterator iter = triggers.iterator(); iter.hasNext(); ) {
				Trigger t = (Trigger) iter.next();
				buf.append(t.getDescription()).append(';');
			}
			return buf.toString();
		}
		
		/**
		 * generate unique key for each trigger from a sequence
		 * @return
		 */
		final String getKey() {
			return UUID.randomUUID().toString();
		}
		
		/**
		 * enumerate execution times for this schedule
		 * @author biyu
		 */
		final class ExecutionTimeEnumeration implements Enumeration {
		
			private Date currentExecutionTime = null;
			private Date nextExecutionTime = null;
		
			ExecutionTimeEnumeration() { }
		
			public boolean hasMoreElements() {
				Date d = null;
				for (Iterator iter = triggers.iterator(); iter.hasNext();) {
					Trigger t = (Trigger) iter.next();
					// going forward
					if (currentExecutionTime != null) {
						d = t.getFireTimeAfter(currentExecutionTime);
					} else {
						d = t.getNextFireTime();
					}
				}
				if (d != null) {
					nextExecutionTime = d;
					return true;
				}
				return false;
			}
		
			public Object nextElement() {
				currentExecutionTime = nextExecutionTime;
				return currentExecutionTime;
			}
		}

		public int getCronMisFireInstruction() {
			return cronMisfireInstruction;
		}

		public int getSimpleMisFireInstruction() {
			return simpleMisfireInstruction;
		}

		public void setCronMisFireInstruction(int misfireInstruction) {
			this.cronMisfireInstruction = misfireInstruction;
			for (Iterator iter = triggers.iterator(); iter.hasNext();) {
				Trigger trigger = (Trigger) iter.next();
				if (trigger instanceof CronTrigger) {
					trigger.setMisfireInstruction(misfireInstruction);
				}
			}
		}

		public void setSimpleMisFireInstruction(int misfireInstruction) {
			this.simpleMisfireInstruction = misfireInstruction;
			for (Iterator iter = triggers.iterator(); iter.hasNext();) {
				Trigger trigger = (Trigger) iter.next();
				if (trigger instanceof SimpleTrigger) {
					trigger.setMisfireInstruction(misfireInstruction);
				}
			}
		}
	}

	/**
	 * @see org.lightj.schedule.SchedulerFactory#getSchedules(java.lang.Class)
	 */
	public List<ISchedule> getSchedules(Schedulable schedulable) throws ScheduleException {
		List<String> ids = new ArrayList<String>();
		ids.add(schedulable.getKey());
		return getSchedules(ids, schedulable.getClass().getName());
	}
	
	/**
	 * find all quartz schedules based on a list locatable string
	 * @param ids
	 * @return
	 */
	private List<ISchedule> getSchedules(List<String> ids, String groupName) {
		List<ISchedule> rst = new ArrayList<ISchedule>();
		if (ids.isEmpty()) return rst;
		try {
			String[] jobNames = scheduler.getJobNames(groupName);
			for (int i = 0, len = jobNames.length; i < len; i++) {
				if (jobNames[i] != null && jobNames[i].indexOf('.') > 0 && 
						ids.contains(jobNames[i].substring(jobNames[i].indexOf('.')+1))) 
				{
					JobDetail job = scheduler.getJobDetail(jobNames[i], groupName);
					if (job != null) {
						QuartzSchedule schedule = new QuartzSchedule(job);
						rst.add(schedule);					
					}
				}
			}
		}
		catch (SchedulerException e) {
			// can't find job, proceed
			_log.error("failed to search schedules", e);
		}
		return rst;
	}
	
	@Override
	public XMLElement getStatsXml() {
		XMLElement schElem = new XMLElement();
		schElem.setName("Scheduler");
		try {
			for (String group : scheduler.getJobGroupNames()) {
				for (String jobname : scheduler.getJobNames(group)) {
					JobDetail job = scheduler.getJobDetail(jobname, group);
					XMLElement jobElem = new XMLElement();
					jobElem.setName("Job");
					jobElem.setAttribute("fullName", job.getFullName());
					jobElem.setAttribute("dscription", job.getDescription());
					jobElem.setAttribute("group", job.getGroup());
					jobElem.setAttribute("metadata", StringUtil.join(job.getJobDataMap().getWrappedMap(), ","));
					Trigger[] triggers = scheduler.getTriggersOfJob(jobname, group);
					if (triggers != null && triggers.length>0) {
						jobElem.setAttribute("triggerSize", triggers.length);
						XMLElement tElem = new XMLElement();
						tElem.setName("Trigger");
						for (Trigger t : triggers) {
							tElem.setAttribute("startTime", t.getStartTime()!=null ? fmt.format(t.getStartTime()) : "n/a");
							tElem.setAttribute("endTime", t.getEndTime()!=null ? fmt.format(t.getEndTime()) : "n/a");
							tElem.setAttribute("nextFireTime", t.getNextFireTime()!=null ? fmt.format(t.getNextFireTime()) : "n/a");
							tElem.setAttribute("prevFireTime", t.getPreviousFireTime()!=null ? fmt.format(t.getPreviousFireTime()) : "n/a");
							jobElem.addChild(tElem);
						}
					}
					schElem.addChild(jobElem);
				}
			}
		} catch (SchedulerException e) {
			schElem.setAttribute("error", e.getMessage());
		}
		return schElem;
	}

}
