package org.lightj.schedule;

import org.lightj.RuntimeContext;
import org.lightj.dal.BaseDatabaseType;
import org.lightj.initialization.BaseInitializable;
import org.lightj.initialization.BaseModule;
import org.lightj.initialization.InitializationException;
import org.lightj.util.Log4jProxy;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.impl.DirectSchedulerFactory;

/**
 * place to intialize your scheduler, take array of {@link SchedulerTypeEnum}
 * 
 * {@link SchedulerTypeEnum#SCHEDULER_PERSISTENT} persistent scheduler 
 * need a config file quartz.persistent.properties in its config root {@link RuntimeContext#getConfigRoot()}
 * persistent storage set by {@link SchedulerConfigBean#setPersistentSchedulerDb(BaseDatabaseType)} prior to initialize this module 
 * 
 * {@link SchedulerTypeEnum#SCHEDULER_TRANSIENT} transient scheduler
 * scheduler threadpool size can be set by {@link SchedulerConfigBean#setTransientSchedulerThreadSize(int)} prior to initialize this module
 * 
 * @author biyu
 *
 */
public class TransientSchedulerModule {
	
	/** logger */
	static Log4jProxy logger = Log4jProxy.getInstance(TransientSchedulerModule.class);

	/** singleton */
	private static TransientSchedulerModuleInner s_Module = null;
	
	public TransientSchedulerModule() {
		init();
	}
	
	private synchronized void init() {
		if (s_Module == null) s_Module = new TransientSchedulerModuleInner();
	}
	
	public BaseModule getModule() {
		return s_Module;
	}

	private class TransientSchedulerModuleInner extends BaseModule {
		
		/** constructor, initialize */
		private TransientSchedulerModuleInner() {
			super(new BaseModule[] {});
			addInitializable(new BaseInitializable() {
				
				@Override
				protected void initialize() {

					try {
						initTransientScheduler(5);
					} catch (ScheduleException e) {
						logger.fatal("Scheduler failed to initialize", e);
						throw new InitializationException(e);
					}
				}

				@Override
				protected void shutdown() {
					try {
						SchedulerFactory.getInstance(SchedulerTypeEnum.SCHEDULER_TRANSIENT).shutdown();
					} catch (ScheduleException e) {
						logger.error("Failed to shutdown scheduler " + SchedulerTypeEnum.SCHEDULER_TRANSIENT, e);
					}
				}
				
			});
		}

	}
	
	
	/**
	 * instantiate trainsient scheduler
	 * @param threadPoolSize
	 * @throws ScheduleException
	 */
	private static final void initTransientScheduler(int threadPoolSize) throws ScheduleException {
		try {
			DirectSchedulerFactory.getInstance().createVolatileScheduler(threadPoolSize);
			Scheduler scheduler = DirectSchedulerFactory.getInstance().getScheduler(DirectSchedulerFactory.DEFAULT_SCHEDULER_NAME);
			SchedulerFactory.addInstance(SchedulerFactory.SCHEDULER_TRANSIENT, scheduler, true);
			scheduler.start();
			logger.info("Transient scheduler started...");
			SchedulerFactory.getInstance(SchedulerFactory.SCHEDULER_TRANSIENT).start();
		} catch (SchedulerException e) {
			throw new ScheduleException(e);
		}
	}
	
}

