package org.lightj.session;

import java.io.InvalidObjectException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.lightj.session.dal.ISessionMetaData;
import org.lightj.session.dal.SessionDataFactory;
import org.lightj.session.step.StepProgressInfo;
import org.lightj.session.step.StepProgressInfo.ProgressState;
import org.lightj.util.DateUtil;
import org.lightj.util.Log4jProxy;
import org.lightj.util.StringUtil;


/**
 * session manager context, store context for a workflow session, support property of String type
 * and Blob type in a name value pair format
 * 
 * @author biyu
 */
public class FlowContext {
	
	static final Log4jProxy cat = Log4jProxy.getInstance(FlowContext.class);
	static final String dateFormat = "yyyy-MM-dd kk:mm:ss.SSS";
	
	/** call back property used for session recovery */
	public static final String PROPERTY_TASKID 		= 	"PROPERTY_TASKID";
	
	public static final String PROPERTY_STEP_PROGRESS	=	"PROPERTY_STEP_PROGRESS";
	
	/**
	 * property hash map
	 */
	private Map<String, ISessionMetaData> context = new LinkedHashMap<String, ISessionMetaData>();
	
	/**
	 * flag for lazy loading
	 */
	private volatile boolean loaded = false;
	
	/**
	 * session id this context associated with
	 */
	private long sessionId;
	
	/**
	 * Constructor
	 * @param sm
	 */
	public FlowContext() {}
	
	/**
	 * set session id
	 * @param sessionId
	 */
	public void setSessionId(long sessionId) {
		this.sessionId = sessionId;
	}
	
	public long getSessionId() {
		return sessionId;
	}
	
	/**
	 * lazy load session context only when called upon
	 *
	 */
	private synchronized void lazyLoad() {
		if (!loaded) {
			if (sessionId > 0) {
				FlowSessionFactory.getInstance().loadMeta(this.sessionId, this);
			}
			loaded = true;
		}
	}
	
	/**
	 * property name set
	 * @return
	 */
	public Set<String> getNames(){
		lazyLoad();
		return context.keySet();
	}
	
	/**
	 * get string property value
	 * @param name
	 * @return
	 */
	public String getStringMeta(String name) {
		lazyLoad();
		if (context != null && context.get(name) != null){
			return context.get(name).getStrValue();	
		}
		return null;
	}

	/**
	 * get blob property value
	 * @param name
	 * @return
	 */
	public Serializable getBlobMeta(String name) {
		lazyLoad();
		if (context != null && context.containsKey(name)) {
			return context.get(name).getBlobValue();
		}
		else {
			return null;
		}
	}
	
	/**
	 * set string property value
	 * @param name
	 * @param value
	 */
	public void setStringMeta(String name, String value) {
		lazyLoad();
		if (context.containsKey(name)) {
			// is update
			context.get(name).setStrValue(value);
			context.get(name).setDirty(true);
		}
		else {
			// new meta
			ISessionMetaData managerMeta = SessionDataFactory.getInstance().getMetaDataManager().newInstance();
			// session id is set when it is being saved
			managerMeta.setName(name);
			managerMeta.setStrValue(value);
			managerMeta.setDirty(true);
			context.put(name, managerMeta);
		}
	}
	
	/**
	 * set blob property value
	 * @param name
	 * @param value
	 */
	public void setBlobMeta(String name, Serializable value) {
		lazyLoad();
		if (context.containsKey(name)) {
			// is update
			context.get(name).setBlobValue(value);
			context.get(name).setDirty(true);
		}
		else {
			ISessionMetaData managerMeta = SessionDataFactory.getInstance().getMetaDataManager().newInstance();
			// session id is set when it is being saved
			managerMeta.setName(name);
			managerMeta.setBlobValue(value);
			managerMeta.setDirty(true);
			context.put(name, managerMeta);
		}
	}
	
	/**
	 * get properties changed since last time saved
	 * @return
	 */
	public List<ISessionMetaData> getDirtyMetas() {
		lazyLoad();
		List<ISessionMetaData> rst = new ArrayList<ISessionMetaData>();
		for (ISessionMetaData meta: context.values()) {
			if (meta.isDirty()) {
				rst.add(meta);
			}
		}
		return rst;
	}
	
	/**
	 * load properties from db
	 * @param metas
	 */
	public void load(List<ISessionMetaData> metas) {
		for (ISessionMetaData meta : metas) {
			if (!StringUtil.isNullOrEmpty(meta.getStrValue()) ||
					meta.getBlobValue() != null) {
				context.put(meta.getName(), meta);
			}
		}
	}
	
	/**
	 * set dirty flag
	 * @param name
	 */
	protected void setDirty(String name, boolean isDirty) {
		if (exist(name)) {
			context.get(name).setDirty(isDirty);
		}
	}
	
	/**
	 * if a property with specified name exists
	 * @param name
	 * @return
	 */
	public boolean exist(String name) {
		lazyLoad();
		return context.containsKey(name);
	}
	
	/////////////////// convenient methods /////////////////
	
	public void setIntParam(String name, int value) {
		setStringMeta(name, Integer.toString(value));
	}
	
	public int getIntParam(String name) throws InvalidObjectException {
		lazyLoad();
		try {
			return Integer.valueOf(getStringMeta(name)).intValue();
		} catch (Throwable t) {
			throw new InvalidObjectException(t.getMessage());
		}
	}
	
	public int getIntParam(String name, int defValue) {
		lazyLoad();
		try {
			return Integer.valueOf(getStringMeta(name)).intValue();
		} catch (Throwable t) {
			return defValue;
		}
	}

	public void setLongParam(String name, long value) {
		setStringMeta(name, Long.toString(value));
	}
	
	public long getLongParam(String name) throws InvalidObjectException {
		lazyLoad();
		try {
			return Long.valueOf(getStringMeta(name)).longValue();
		} catch (Throwable t) {
			throw new InvalidObjectException(t.getMessage());
		}
	}
	
	public long getLongParam(String name, long defValue) {
		lazyLoad();
		try {
			return Long.valueOf(getStringMeta(name)).longValue();
		} catch (Throwable t) {
			return defValue;
		}
	}

	public void setDateParam(String name, Date date) {
		setStringMeta(name, DateUtil.format(date, dateFormat));
	}
	
	public Date getDateParam(String name) throws InvalidObjectException {
		lazyLoad();
		try {
			return DateUtil.parse(getStringMeta(name), dateFormat);
		} catch (Throwable t) {
			throw new InvalidObjectException(t.getMessage());
		}
	}
	
	public Date getDateParam(String name, Date defValue) {
		lazyLoad();
		try {
			return DateUtil.parse(getStringMeta(name), dateFormat);
		} catch (Throwable t) {
			return defValue;
		}
	}

	public void setBooleanParam(String name, boolean value) {
		setStringMeta(name, Boolean.toString(value));
	}
	
	public boolean getBooleanParam(String name) throws InvalidObjectException {
		lazyLoad();
		try {
			return Boolean.valueOf(getStringMeta(name)).booleanValue();
		} catch (Throwable t) {
			throw new InvalidObjectException(t.getMessage());
		}
	}

	public boolean getBooleanParam(String name, boolean defValue) {
		lazyLoad();
		try {
			return Boolean.valueOf(getStringMeta(name)).booleanValue();
		} catch (Throwable t) {
			return defValue;
		}
	}

	public void setStringParam(String name, String value) {
		setStringMeta(name, value);
	}
	
	public String getStringParam(String name) throws InvalidObjectException {
		lazyLoad();
		if (exist(name)) {
			return getStringMeta(name);
		}
		else {
			throw new InvalidObjectException(name + " not exist");
		}
	}

	public String getStringParam(String name, String defValue) {
		lazyLoad();
		if (exist(name)) {
			return getStringMeta(name);
		}
		else {
			return defValue;
		}
	}
	
	public boolean isPropertyPrivate(String propertyName) {
		return false;
	}

	public String getTaskId() {
		return getStringParam(PROPERTY_TASKID, null);
	}
	
	public void setTaskId(String taskId) {
		setStringParam(PROPERTY_TASKID, taskId);
	}
	
	public synchronized void updateStepProgress(StepProgressInfo stepProgress) {
		Serializable sps = getBlobMeta(PROPERTY_STEP_PROGRESS);
		if (sps != null) {
			StepProgressInfo[] spAll = (StepProgressInfo[]) sps;
			int i = 0, len = spAll.length;
			for (; i < len; i++) {
				StepProgressInfo sp = spAll[i];
				// update
				if (StringUtil.equalIgnoreCase(sp.stepName, stepProgress.stepName) &&
						(StringUtil.equalIgnoreCase(sp.executionId, stepProgress.executionId))) {
					spAll[i].state = stepProgress.state;
					if (stepProgress.info != null) {
						spAll[i].info += (stepProgress.info + '\n');
					}
					break;
				}
			}
			// create
			if (i == len) {
				StepProgressInfo[] newSps = new StepProgressInfo[len+1];
				System.arraycopy(spAll, 0, newSps, 0, len);
				newSps[len] = stepProgress;
			}
		}
		else {
			// new array
			StepProgressInfo[] newSps = new StepProgressInfo[1];
			newSps[0] = stepProgress;
			setBlobMeta(PROPERTY_STEP_PROGRESS, newSps);
		}
		setDirty(PROPERTY_STEP_PROGRESS, true);
	}
	
	public synchronized void updateAllStepState(String stepName, ProgressState state) {
		for (StepProgressInfo info : getStepProgress(stepName)) {
			info.state = state;
		}
		setDirty(PROPERTY_STEP_PROGRESS, true);
	}
	
	public StepProgressInfo[] getStepProgress(String stepName) {
		Serializable sps = getBlobMeta(PROPERTY_STEP_PROGRESS);
		ArrayList<StepProgressInfo> rst = new ArrayList<StepProgressInfo>();
		if (sps != null) {
			StepProgressInfo[] spAll = (StepProgressInfo[]) sps;
			for (StepProgressInfo sp : spAll) {
				if (StringUtil.equalIgnoreCase(sp.stepName, stepName)) {
					rst.add(sp);
				}
			}
		}
		return rst.toArray(new StepProgressInfo[0]);
	}
	
}
