package org.lightj.session.step;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import org.lightj.session.FlowContext;
import org.lightj.session.FlowEvent;
import org.lightj.session.FlowSession;
import org.lightj.session.FlowSessionFactory;
import org.lightj.session.IFlowEventListener;
import org.lightj.task.ITaskListener;
import org.lightj.task.Task;
import org.lightj.task.TaskExecutionException;
import org.lightj.task.TaskResult;
import org.lightj.task.TaskResultEnum;
import org.lightj.task.TaskRunner;

@SuppressWarnings("rawtypes")
public class FlowStepBuilder<T extends FlowContext> extends StepBuilder {
	
	public FlowStepBuilder launchSubFlows(
			Collection<FlowSession> subFlows, 
			StepErrorHandler execErrorHandler, 
			StepCallbackHandler resultHandler, 
			StepErrorHandler resultErrorHandler) 
	{
		FlowStepExecution<T> execution = new FlowStepExecution<T>(new FlowTaskRunner(), resultHandler);
		List<FlowTask> tasks = new ArrayList<FlowTask>();
		for (FlowSession subFlow : subFlows) {
			tasks.add(new FlowTask(subFlow));
		}
		execution.addTasks(tasks.toArray(new FlowTask[0]));
		super.execute(execution, execErrorHandler);
		super.onCallback(resultHandler, resultErrorHandler);
		return this;
	}
	
	/**
	 * inner task to run sub flows
	 * @author biyu
	 *
	 */
	static class FlowTask extends Task {

		/**
		 * subflows to be run
		 */
		private FlowSession subFlow;
		
		protected FlowSession getSubFlow() {
			return subFlow;
		}

		private long parentFlowId;
		
		protected void setParentFlowId(long parentFlowId) {
			this.parentFlowId = parentFlowId;
		}

		protected long getParentFlowId() {
			return parentFlowId;
		}

		FlowTask(FlowSession subFlow) {
			super();
			this.subFlow = subFlow;
		}
		
	}
	
	/**
	 * innter task runner to run sub flows
	 * @author biyu
	 *
	 */
	static class FlowTaskRunner extends TaskRunner<FlowTask> {

		private AtomicInteger runningCount = new AtomicInteger(0);
		
		/**
		 * run task business logic, subclass implement this for execution logic
		 * @param task
		 * @return
		 * @throws TaskExecutionException
		 */
		protected String runASyncTaskInternal(ITaskListener taskListener, FlowTask task) throws TaskExecutionException {
			
			FlowSession subFlow = task.getSubFlow();
			try {
				runningCount.incrementAndGet();
				if (subFlow.getParentId() <= 0) {
					subFlow.setParentId(task.getParentFlowId());
					FlowSessionFactory.getInstance().update(subFlow);
				}
				subFlow.addEventListener(new SubFlowEventListener(task, taskListener, runningCount));
				subFlow.runFlow();
			} catch (Throwable t) {
				runningCount.decrementAndGet();
			}
			return subFlow.getKey();
		}

	}
	
	/**
	 * inner flow event listener to call back parent flow on join
	 * @author biyu
	 *
	 */
	static class SubFlowEventListener implements IFlowEventListener {
		
		private Task task;
		private ITaskListener taskListener;
		private AtomicInteger runningCount;
		public SubFlowEventListener(FlowTask task, ITaskListener taskListener, AtomicInteger runningCount) {
			this.task = task;
			this.taskListener = taskListener;
			this.runningCount = runningCount;
		}
		
		@Override
		public void handleStepEvent(FlowEvent event,
				FlowSession session, IFlowStep flowStep,
				StepTransition stepTransition) {
			// noop
		}

		@Override
		public void handleFlowEvent(FlowEvent event, FlowSession session) {
			if (event == FlowEvent.stop) {
				int count = runningCount.decrementAndGet();
				try {
					TaskResultEnum status = session.getResult().toTaskResult();
					Task task = new Task();
					task.setTaskId(session.getKey());
					taskListener.handleTaskResult(TaskResult.createLocal(task, status, session.getStatus()));
				} finally {
					if (count == 0) {
						taskListener.taskCompleted(task);
					}
				}
			}
		}

		@Override
		public void handleError(Throwable t, FlowSession session) {
			// noop
		}		
	}
}
