package org.lightj.session.step;

import java.util.HashMap;

import org.lightj.session.FlowContext;
import org.lightj.session.FlowExecutionException;
import org.lightj.task.Task;
import org.lightj.task.TaskResult;
import org.lightj.task.TaskResultEnum;


/**
 * A convenient call back handler that maps different callback result to different flow transitions
 * the callback happened on each task result generation, ie. on handleTaskResult, the default behavior of
 * taskCompleted() is noop
 * 
 * @author biyu
 *
 */
public class MappedCallbackHandler<T extends FlowContext> extends StepCallbackHandler<T> {
	
	/**
	 * map a result status to a flow step result
	 */
	protected HashMap<TaskResultEnum, StepTransition> mapOnResults = new HashMap<TaskResultEnum, StepTransition>();
	
	/**
	 * Construct a mapped callback handler with default transition
	 * @param defResult
	 */
	public MappedCallbackHandler(StepTransition defResult) {
		super(defResult);
	}
	
	/**
	 * register a status with a result
	 * @param status
	 * @param result
	 */
	public void mapResultTo(TaskResultEnum status, StepTransition result) {
		mapOnResults.put(status, result);
	}
	
	/**
	 * register statuses with results
	 * @param mapOnResults
	 */
	public void mapResultsTo(HashMap<TaskResultEnum, StepTransition> mapOnResults) {
		this.mapOnResults.putAll(mapOnResults);
	}
	
	@Override
	public synchronized StepTransition executeOnCompleted(Task task)
			throws FlowExecutionException 
	{
		StepTransition transition = defResult;
		// if we want to move on when result comes back, find the next step
		TaskResult curRst = null;
		for (TaskResult result : results) {
			TaskResultEnum status = result.getStatus();
			if (mapOnResults.containsKey(status) && 
					(curRst == null || result.isMoreSevere(curRst))) {
				curRst = result;
				transition = mapOnResults.get(status);
				transition.log(result.getMsg());
			}
		}
		return transition;
	}

	/**
	 * log and change session context
	 */
	@Override
	public StepTransition executeOnSubmitted(Task task) throws FlowExecutionException {
		StepTransition transition = StepTransition.newLog(task.getTaskId());
		return transition;
	}

	/**
	 * logging
	 */
	@Override
	public StepTransition executeOnCreated(Task task) throws FlowExecutionException {
		return StepTransition.NOOP;
	}

	/**
	 * move the flow to the transitions predefined in the result to transition map,
	 * or to default transition if nothing matches
	 */
	@Override
	public synchronized StepTransition executeOnResult(TaskResult result) throws FlowExecutionException {
		// remember result
		return StepTransition.newLog((result.getStatus() + ": " + result.getMsg()));
	}

}
