package org.lightj.session.step;

import java.util.LinkedHashMap;
import java.util.Map;

import org.lightj.session.FlowContext;
import org.lightj.session.FlowExecutionException;
import org.lightj.util.StringUtil;


/**
 * A convenient error handler that map different Throwable class type to different transitions
 * the map is ordered, transition will be picked by the first throwable class from the map
 * that is assignable to the thrown exception. The behavior is much like java catch block
 * 
 *  } catch (e1) {
 *  } catch (e2) {
 *  ...
 * 
 * @author biyu
 *
 */
public class MappedErrorHandler<T extends FlowContext> extends StepErrorHandler<T> {
	
	public MappedErrorHandler(StepTransition transition) {
		super(transition);
	}

	/**
	 * map an error type to a result
	 */
	protected Map<Class<? extends Throwable>, StepTransition> errorClass2ResultMap = new LinkedHashMap<Class<? extends Throwable>, StepTransition>();

	/**
	 * register an error class to a result
	 * @param errorKlass
	 * @param result
	 */
	public void mapErrorTo(Class<? extends Throwable> errorKlass, StepTransition result) {
		errorClass2ResultMap.put(errorKlass, result);
	}
	
	/**
	 * register an error class to a result
	 * @param errorKlass
	 * @param result
	 */
	public void mapErrorsTo(Map<Class<? extends Throwable>, StepTransition> errorClass2ResultMap) {
		this.errorClass2ResultMap.putAll(errorClass2ResultMap);
	}

	@Override
	public StepTransition executeOnError(Throwable t) {
		Throwable th = null;
		StepTransition tran = defResult;
		if (t instanceof FlowExecutionException && t.getCause() != null) {
			// checked exception
			th = t.getCause();
		}
		else {
			// runtime exception
			th = t;
		}
		
		Class<? extends Throwable> errorKlass = th.getClass();
		for (Class<? extends Throwable> mappedKlass : errorClass2ResultMap.keySet()) {
			if (mappedKlass.isAssignableFrom(errorKlass)) {
				tran = errorClass2ResultMap.get(mappedKlass);
				break;
			}
		}

		if (StringUtil.isNullOrEmpty(th.getMessage())) {
			tran.log(StringUtil.trimToLength(StringUtil.getStackTrace(th),2000));
		}
		else {
			tran.log(th.getMessage());
		}
		return tran;
	}

}
