package org.lightj.session.step;

import java.util.ArrayList;
import java.util.List;

import org.lightj.session.FlowContext;
import org.lightj.session.FlowExecutionException;
import org.lightj.session.FlowResult;
import org.lightj.session.step.StepProgressInfo.ProgressState;
import org.lightj.task.Task;
import org.lightj.task.TaskResult;
import org.lightj.util.Log4jProxy;


/**
 * An execution aspect of a flow step to handle asynchronous callback
 * 
 * @author biyu
 *
 */
public abstract class StepCallbackHandler<T extends FlowContext> extends StepExecution<T> {
	
	static Log4jProxy logger = Log4jProxy.getLogger(StepCallbackHandler.class.getName());

	/**
	 * result when the task is created
	 */
	public static final int TYPE_CREATED		=	1;
	
	/**
	 * result when the task is submitted
	 */
	public static final int TYPE_SUBMITTED		=	2;
	
	/**
	 * result when the task result is available
	 */
	public static final int TYPE_TASKRESULT		=	3;

	/**
	 * result when the task is completed
	 */
	public static final int TYPE_COMPLETED		=	4;
		
	/**
	 * result type
	 */
	protected volatile int callbackType;
	
	/**
	 * task result
	 */
	protected List<TaskResult> results = new ArrayList<TaskResult>();
	
	/**
	 * task to be executed
	 */
	protected Task task;
	
	/**
	 * construct a result handler with default transition
	 * @param transition
	 */
	public StepCallbackHandler(StepTransition transition) {
		super(transition);
	}
	
	/**
	 * result handle task completed event
	 */
	public final void handleTaskResult(TaskResult result) {
		callbackType = TYPE_TASKRESULT;
		if (result != null) {
			this.results.add(result);
		}
		resumeFlowStep();
	}

	/**
	 * result handle task submitted event
	 */
	public final void taskSubmitted(Task task) {
		callbackType = TYPE_SUBMITTED;
		if (this.task == null) {
			this.task = task;
		}
		resumeFlowStep();
	}

	/**
	 * result handle task created event
	 */
	public final void taskCreated(Task task) {
		callbackType = TYPE_CREATED;
		if (this.task == null) {
			this.task = task;
		}
		resumeFlowStep();
	}

	/**
	 * result handle task completion event
	 */
	public final void taskCompleted(Task task) {
		callbackType = TYPE_COMPLETED;
		if (this.task == null) {
			this.task = task;
		}
		resumeFlowStep();
	}

	@Override
	public final StepTransition execute() throws FlowExecutionException {
		switch (callbackType) {
		case TYPE_SUBMITTED:
			saveStepProgress(task);
			return executeOnSubmitted(task);
		case TYPE_TASKRESULT:
			TaskResult result = results.get(results.size()-1);
			saveStepProgress(ProgressState.Result, result.getMsg());
			return executeOnResult(result);
		case TYPE_CREATED:
			return executeOnCreated(task);
		case TYPE_COMPLETED:
			saveStepProgress(ProgressState.Complete, "");
			return executeOnCompleted(task);
		}
		throw new FlowExecutionException("Invalid callback type " + callbackType);
	}
	
	/** convert {@link ICallableResult} status to {@link FlowResult} */
	protected FlowResult convertStatus(TaskResult result) {
		if (result.isSuccess()) return FlowResult.Success;
		else if (result.isFailed()) return FlowResult.Failed;
		else if (result.isTimedOut()) return FlowResult.Timeout;
		else if (result.isCancelled()) return FlowResult.Canceled;
		else return FlowResult.Unknown;
	}

	/** 
	 * save task progress
	 */
	protected void saveStepProgress(Task task) {
		StepProgressInfo sp = StepProgressInfo.newInstance(
				flowStep.getStepName(), 
				getExecutionId(), 
				ProgressState.Callback, 
				task.getTaskId());
		sessionContext.updateStepProgress(sp);
	}

	/**
	 * save task progress
	 * @param result
	 */
	protected void saveStepProgress(ProgressState state, String msg) {
		StepProgressInfo sp = StepProgressInfo.newInstance(
				flowStep.getStepName(), getExecutionId(), 
				state, 
				msg);
		sessionContext.updateStepProgress(sp);
	}

	/**
	 * do the work when task completed
	 * @param result
	 * @return
	 * @throws FlowExecutionException
	 */
	public abstract StepTransition executeOnCompleted(Task task) throws FlowExecutionException;
	
	/**
	 * do the work when task submitted
	 * @param task
	 * @return
	 * @throws FlowExecutionException
	 */
	public abstract StepTransition executeOnSubmitted(Task task) throws FlowExecutionException;
	
	/**
	 * do the work when task is created
	 * @param task
	 * @return
	 * @throws FlowExecutionException
	 */
	public abstract StepTransition executeOnCreated(Task task) throws FlowExecutionException;
	
	/**
	 * do the work when task generates some result
	 * @param result
	 * @return
	 * @throws FlowExecutionException
	 */
	public abstract StepTransition executeOnResult(TaskResult result) throws FlowExecutionException;
	
}
