package org.lightj.session.step;

import org.lightj.session.FlowResult;
import org.lightj.session.FlowState;

/**
 * A transition from one flow step to another, this represents the arch in flow chart from
 * one state to another
 * 
 * possible combinations and their meaning:
 * 
 * actionstatus, nextstep, resultstatus, msg	: meaning
 * 1. running, nextstep, *, * 	: run next step
 * 2. callback, null, *, * 		: park flow, waiting for callback
 * 3. null, null, *, *			: don't change flow state, just change result or do logging
 * 4. completed/canceled/skipped, *, *, *		: stop flow, close session
 * 5. waiting/retry/suspended, nextstep, *, *		: park flow at a step, waiting for external action
 *  
 * @author biyu
 *
 */
@SuppressWarnings("rawtypes")
public final class StepTransition {
	
	public static final StepTransition CALLBACK	=	new StepTransition().inState(FlowState.Callback);
	public static final StepTransition NOOP = new StepTransition();
	
	/**
	 * flow activity state
	 */
	private FlowState actionStatus;
	
	/**
	 * flow result state
	 */
	private FlowResult resultStatus;
	
	/**
	 * next flow step to transition to
	 */
	private String nextStep;
	
	/**
	 * message for the flow execution
	 */
	private String msg;

	/** a new log instance */
	public static StepTransition newLog(String msg) {
		return new StepTransition().log(msg);
	}
	
	/** run to a step */
	public static StepTransition runToStep(String stepName) {
		return new StepTransition().toStep(stepName).inState(FlowState.Running);
	}

	/** run to a step */
	public static StepTransition runToStep(Enum stepName) {
		return new StepTransition().toStep(stepName).inState(FlowState.Running);
	}

	public StepTransition() {}
	
	/**
	 * construct a FlowStepTransition given the destination step
	 * flow state is default to {@link FlowState#Running}
	 * @param nextStep
	 */
	public StepTransition(String nextStep) {
		this(FlowState.Running, nextStep, null, null);
	}
	
	/**
	 * constructor
	 * @param nextStep
	 */
	public StepTransition(Enum nextStep) {
		this(FlowState.Running, nextStep.name(), null, null);
	}

	/**
	 * construct a FlowStepTransition given a flow state and next step it goes to
	 * @param actionStatus
	 * @param nextStep
	 */
	public StepTransition(FlowState actionStatus, String nextStep) {
		this(actionStatus, nextStep, null, null);
	}
	
	/**
	 * construct a FlowStepTransition given a flow state and next step it goes to
	 * @param actionStatus
	 * @param nextStep
	 */
	public StepTransition(FlowState actionStatus, Enum nextStep) {
		this(actionStatus, nextStep.name(), null, null);
	}

	/**
	 * construct a {@link StepExecution} given a flow state, next step, result state and message
	 * @param actionStatus
	 * @param nextStep
	 * @param resultStatus
	 * @param msg
	 */
	public StepTransition(FlowState actionStatus, String nextStep, FlowResult resultStatus, String msg) {
		this.actionStatus = actionStatus;
		this.nextStep = nextStep;
		this.resultStatus = resultStatus;
		this.msg = msg;
	}
	
	public StepTransition inState(FlowState state) {
		this.actionStatus = state;
		return this;
	}

	public StepTransition toStep(String stepName) {
		this.nextStep = stepName;
		this.actionStatus = FlowState.Running;
		return this;
	}

	public StepTransition toStep(Enum stepName) {
		this.nextStep = stepName.name();
		this.actionStatus = FlowState.Running;
		return this;
	}

	public StepTransition withResult(FlowResult result) {
		this.resultStatus = result;
		return this;
	}

	public StepTransition log(String msg) {
		this.msg = msg;
		return this;
	}

	public FlowState getActionStatus() {
		return actionStatus;
	}

	public String getNextStep() {
		return nextStep;
	}

	public String getMsg() {
		return msg;
	}

	public FlowResult getResultStatus() {
		return resultStatus;
	}
	
	public boolean isEdge() {
		return ((actionStatus==FlowState.Running && nextStep != null) || 
				(actionStatus!=null && actionStatus.isComplete()));
	}
	
}
