package org.lightj.task;

import java.util.HashSet;
import java.util.concurrent.DelayQueue;

import org.lightj.concurrent.AbstractQueueTask;
import org.lightj.concurrent.CompositeBlockingQueue;
import org.lightj.concurrent.PriorityRoundRobinThreadPoolExecutor;
import org.lightj.concurrent.QueueTaskGroups;
import org.lightj.concurrent.QueueTaskPriority;
import org.lightj.util.Log4jProxy;


/**
 * Monitor external task running
 * 
 * @author biyu
 */
@SuppressWarnings({"rawtypes", "unchecked"})
public class MonitorRunner implements Runnable {
	
	/** logger */
	static Log4jProxy cat = Log4jProxy.getInstance(MonitorRunner.class);
	
	/** singleton */
	private static MonitorRunner me = null;
	
	/** internal queue */
	private CompositeBlockingQueue<MonitorEvent> eventQueue = 
		new CompositeBlockingQueue<MonitorEvent>("MonitorQueue", new DelayQueue<MonitorEvent>());
	
	/**
	 * Singleton
	 * @return singleton
	 */
	public synchronized static MonitorRunner getInstance() {
		if (me == null) {
			me = new MonitorRunner();
			me.start();
		}
		return me;
	}
	
	/** singleton */
	private MonitorRunner() {}

	/**
	 * Process awaiting events in the queue
	 */
	private final void processQueue() {
		while (true) {
			// we don't want the daemon process to be interruptted by any error, or be interrupted
			// this is run from a daemon thread, VM will shutdown anyway if all remaining threads are daemon threads
			try {
				MonitorEvent event = eventQueue.take();
				PriorityRoundRobinThreadPoolExecutor.getInstance().submit(
						event.isTimedOut() ? new TimeoutTask(event) : new MonitorTask(event));
			} catch (InterruptedException e) {	
				// someone has request this thread to be shutdown explicitly, probably the thread pool is shutting down 
			} catch (Throwable t) {
				// some other unknown/runtime error, ignore by this daemon thread
			}
		}
	}
	
	/**
	 * Add an event to be monitored, it will never block since the internal queue is unbounded
	 * @param event
	 */
	public final void addToQueue(MonitorEvent event) {
		eventQueue.offer(event);
	}
	
	/**
	 *  start event processor
	 */
	final synchronized void start() {
		Thread t = new Thread(this);
		t.setDaemon(true);
		t.start();
		Runtime.getRuntime().addShutdownHook(
				new Thread(
					new Runnable() {

						public void run() {
							shutdown();
						}
						
					}));
	}
	
    /**
     * waiting on the queue
     */
	public void run() {
		processQueue();
	}
	
	/**
	 * shutdown the queue
	 *
	 */
	synchronized void shutdown() {
		HashSet<MonitorEvent> events = new HashSet<MonitorEvent>();
		eventQueue.drainTo(events);
		for (MonitorEvent event : events) {
			PriorityRoundRobinThreadPoolExecutor.getInstance().submit(new CancelTask(event));
		}
	}
	
    /** 
     * Monitor Event QueueTask
     * We always give highest priority for handling asynchronous callback,
     * the reason being allow any tasks already in fly to be completed asap.
     * @author biyu
     */
    static class MonitorTask extends AbstractQueueTask implements Runnable {
    	private final MonitorEvent event;
		MonitorTask(MonitorEvent event) {
			super(QueueTaskGroups.TASKMONITOR, event.getName(), QueueTaskPriority.high);
			this.event = event;
		}

		public void run() {
			try {
				boolean needReschedule = event.getTaskMonitor().monitorTask(event);
				if (needReschedule) {
					event.scheduleNextRun();
					MonitorRunner.getInstance().addToQueue(event);
				}
				else {
					event.getTaskListener().taskCompleted(event.getTask());
				}
			} catch (TaskExecutionException e) {
				cat.error(null, e);
			}
		}
    }

    /** 
     * Timeout Event QueueTask
     * We always give highest priority for handling asynchronous callback,
     * the reason being allow any tasks already in fly to be completed asap.
     * @author biyu
     */
    static class TimeoutTask extends AbstractQueueTask implements Runnable {
    	private final MonitorEvent event;
    	TimeoutTask(MonitorEvent event) {
			super(QueueTaskGroups.TASKMONITOR, event.getName(), QueueTaskPriority.high);
			this.event = event;
		}

		public void run() {
			try {
				event.getTaskMonitor().handleTimeout(event);
				event.getTaskListener().handleTaskResult(TaskResult.createLocal(event.getTask(), TaskResultEnum.Timeout, "Task timeout"));
				event.getTaskListener().taskCompleted(event.getTask());
			} catch (TaskExecutionException e) {
				cat.error(null, e);
			}
		}
    }

    /**
     * Cancel Event QueueTask
     * We always give highest priority for handling asynchronous callback,
     * the reason being allow any tasks already in fly to be completed asap.
     * @author biyu
     */
    static class CancelTask extends AbstractQueueTask implements Runnable {
    	private final MonitorEvent event;
		CancelTask(MonitorEvent event) {
			super(QueueTaskGroups.TASKMONITOR, event.getName(), QueueTaskPriority.high);
			this.event = event;
		}

		public void run() {
			try {
				event.getTaskMonitor().handleCancel(event, "Server shutting down");
				event.getTaskListener().handleTaskResult(TaskResult.createLocal(event.getTask(), TaskResultEnum.Canceled, "Task cancelled: Server shutting down"));
				event.getTaskListener().taskCompleted(event.getTask());
			} catch (TaskExecutionException e) {
				cat.error(null, e);
			}
		}
    }

}
