/*
 * Created on Feb 23, 2006
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package org.lightj.task;

import java.io.InvalidObjectException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Timer;
import java.util.TimerTask;

/**
 * task runner, to run sync, asynchronous task [with polling]
 * 
 * @author biyu
 *
 */
public abstract class TaskRunner<T extends Task> {
	
	/**
	 * @param factory
	 */
	public TaskRunner() {}
	
	/**
	 * run task, subclass implement this with the execution logic
	 * @param taskListener
	 * @param task
	 * @throws InvalidObjectException
	 */
	public final TaskResult runSyncTask(T task) throws TaskExecutionException
	{
		throw new UnsupportedOperationException("not implemented");
	}

	/**
	 * run task
	 * @param taskListener
	 * @param task
	 * @throws InvalidObjectException
	 */
	public final void runASyncTask(ITaskListener taskListener, final T task, ExecuteOption options) throws TaskExecutionException
	{
		if (options != null && options.getTimeoutInMs() > 0) {
			Timer timeoutTimer = new Timer();
			final ITaskListener tl = (ITaskListener) Proxy.newProxyInstance(ITaskListener.class.getClassLoader(),
					new Class<?>[] { ITaskListener.class },
					new AsyncTaskListenerTimeoutWrapper(taskListener, timeoutTimer));
			timeoutTimer.schedule(new TimerTask() {
				
				@Override
				public void run() {
					tl.handleTaskResult(TaskResult.createLocal(task, TaskResultEnum.Timeout, "task timed out"));
					tl.taskCompleted(task);
				}
				
			}, options.getTimeOutAt());
			
			taskListener = tl;
		}
		
		// task created callback
		taskListener.taskCreated(task);
		
		if (options != null && options.getInitialDelayMs() > 0) {
			try {
				Thread.sleep(options.getInitialDelayMs());
			} catch (InterruptedException e) {
				// ignore, continue
			}
		}

		String taskId = runASyncTaskInternal(taskListener, task);
		
		// task submitted callback
		task.setTaskId(taskId);
		
		// callback 
		taskListener.taskSubmitted(task);
		
	}

	/**
	 * run task
	 * @param taskListener
	 * @param task
	 * @throws InvalidObjectException
	 */
	public final void runASyncPollTask(ITaskListener taskListener, T task, TaskMonitor<T> taskMonitor) throws TaskExecutionException
	{
		if (taskMonitor == null) throw new IllegalArgumentException("task monitor cannot be empty for async poll task");
		
		// task created callback
		taskListener.taskCreated(task);

		String taskId = runASyncPollTaskInternal(taskListener, task, taskMonitor);

		// task submitted callback
		task.setTaskId(taskId);
		
		// callback 
		taskListener.taskSubmitted(task);
		
		// schedule monitor event
		taskMonitor.scheduleMonitor(taskListener, task);

	}

	/**
	 * run task business logic, subclass implement this for execution logic
	 * @param task
	 * @return
	 * @throws TaskExecutionException
	 */
	protected String runASyncTaskInternal(ITaskListener taskListener, T task) throws TaskExecutionException {
		throw new UnsupportedOperationException("not implemented");
	}

	/**
	 * run task business logic, subclass implements this for execution logic
	 * @param task
	 * @return
	 * @throws TaskExecutionException
	 */
	protected String runASyncPollTaskInternal(ITaskListener taskListener, T task, TaskMonitor<T> taskMonitor) throws TaskExecutionException {
		throw new UnsupportedOperationException("not implemented");
	}
	
	
	private static class AsyncTaskListenerTimeoutWrapper implements InvocationHandler {
		
		private Object realImpl;
		private Timer timeoutTimer;

		public AsyncTaskListenerTimeoutWrapper(Object impl, Timer timeoutTimer) {
			this.realImpl = impl;
			this.timeoutTimer = timeoutTimer;
		}

		@Override
		public Object invoke(Object proxy, final Method method, final Object[] args)
				throws Throwable {
			if ("taskCompleted".equals(method.getName()) && timeoutTimer != null) {
				timeoutTimer.cancel();
			}
			return method.invoke(realImpl, args);
		}	
	}
}
