package org.lightj.task.schedule;

import java.io.InvalidObjectException;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.lightj.dal.Locator;
import org.lightj.dal.LocatorUtil;
import org.lightj.schedule.IExecutionContext;
import org.lightj.schedule.ISchedule;
import org.lightj.schedule.Schedulable;
import org.lightj.schedule.ScheduleException;
import org.lightj.schedule.SchedulerFactory;
import org.lightj.schedule.SchedulerTypeEnum;
import org.lightj.task.ITaskListener;
import org.lightj.task.TaskExecutionException;
import org.lightj.task.TaskResult;
import org.lightj.task.TaskResultEnum;
import org.lightj.task.TaskRunner;
import org.lightj.util.Log4jProxy;


/**
 * run a scheduled task
 * @author biyu
 *
 */
public class ScheduledTaskRunner extends TaskRunner<ScheduledTask> {
	
	/** logger */
	static Log4jProxy logger = Log4jProxy.getInstance(ScheduledTaskRunner.class);
	
	/**
	 * run task business logic, subclass implement this for execution logic
	 * @param task
	 * @return
	 * @throws TaskExecutionException
	 */
	@Override
	protected String runASyncTaskInternal(ITaskListener taskListener, ScheduledTask task) throws TaskExecutionException {
		// setup the callback
		SchedulerFactory factory = SchedulerFactory.getInstance(SchedulerTypeEnum.SCHEDULER_TRANSIENT);

		Schedulable schedulable = new TaskScheduable(taskListener, (ScheduledTask) task);
	
		try {
			factory.schedule(schedulable);
		} catch (ScheduleException e) {
			logger.error(null, e);
			throw new TaskExecutionException(e.getMessage());
		}
		
		return schedulable.getKey();
	}

	/**
	 * locator 
	 * @author biyu
	 *
	 */
	public static class TaskLocator implements Locator<TaskScheduable> {
		
		static Map<String, TaskScheduable> runningTasks = new HashMap<String, TaskScheduable>();
		
		static TaskLocator me = new TaskLocator();
		private TaskLocator() {}
		public static final TaskLocator getInstance() {
			return me;
		}
		
		@Override
		public TaskScheduable findByKey(String key) throws InvalidObjectException {
			String realKey = LocatorUtil.getLocatableString(this.getClass(), key);
			return runningTasks.get(realKey);
		}
		
		static void addTask(TaskScheduable task) {
			String key = LocatorUtil.getLocatableString(task);
			runningTasks.put(key, task);
		}
		static void removeTask(TaskScheduable task) {
			String key = LocatorUtil.getLocatableString(task);
			runningTasks.remove(key);
		}
		
	}

	/**
	 * wrapper with callback logic
	 * @author biyu
	 *
	 */
	static class TaskScheduable implements Schedulable {
		
		private ITaskListener taskListener;
		private ScheduledTask task;
		private Schedulable s;
		
		TaskScheduable(ITaskListener taskListener, ScheduledTask task) {
			this.taskListener = taskListener;
			this.task = task;
			this.s = task.getSchedulable();
			TaskLocator.addTask(this);
		}
		
		/**
		 * Execute a job.
		 * @param ctx	persistent data for this job
		 */
		public synchronized void execute(IExecutionContext ctx) 
		{
			s.execute(ctx);
			TaskResult result = TaskResult.createLocal(task, TaskResultEnum.Success, "scheduled task callback");
			taskListener.handleTaskResult(result);
			boolean expired = true;
			for (ISchedule sch : s.getSchedules()) {
				// add 100 ms when checking to prevent race condition, realistically not possible to schedule a job every 100ms
				expired &= (sch.getFinalExecutionTime().before(new Date(System.currentTimeMillis()+100L)));
			}
			System.out.println("expired? " + expired);
			if (expired) {
				try {
					taskListener.taskCompleted(task);
				} finally {
					TaskLocator.removeTask(TaskScheduable.this);
				}
			}
		}

		@Override
		public void addSchedule(ISchedule schedule) {
			throw new UnsupportedOperationException("Schedule is immutable");
		}

		@Override
		public List<ISchedule> getSchedules() {
			return s.getSchedules();
		}

		@Override
		public void interrupt() {
			s.interrupt();
		}

		@Override
		public void removeSchedule(ISchedule schedule) {
			throw new UnsupportedOperationException("Schedule is immutable");
		}

		@Override
		public String getKey() {
			return s.getKey();
		}

		@SuppressWarnings("rawtypes")
		@Override
		public Class<? extends Locator> getLocator() {
			return TaskLocator.class;
		}

	}

}
